
class Dielectric:
	def __init__(self, constant, loss) -> None:
		self._Constant =constant
		self._Loss = loss
	
	@property
	def constant(self):
		pass

	@constant.getter
	def constant(self):
		return self._Constant
	
	@property
	def loss(self):
		pass

	@loss.getter
	def loss(self):
		return self._Loss



import types as _types
import math


class FiniteDiff:
	"""
	Computes the nth derivative of a function using forward, backward or central
	finite difference methods 
	"""
	def __init__(self, f:_types.FunctionType, x:float, dx=1E-4, n=1) -> None:
		assert dx>0, "dx>0 expected"
		self._func = f
		self._x = x
		self._dx = dx
		self._n = n


	def forward(self)->float:
		f = self._func
		x = self._x
		dx = self._dx
		n = self._n

		numerator = 0
		for i in range(n+1):
			numerator += (-1)**(n-i)*math.comb(n, i)*f(x+i*dx)
		
		return numerator/dx**n

		
	def backward(self)->float:
		f = self._func
		x = self._x
		dx = self._dx
		n = self._n

		numerator = 0
		for i in range(n+1):
			numerator += (-1)**(i)*math.comb(n, i)*f(x-i*dx)
		
		return numerator/dx**n


	def central(self)->float:
		f = self._func
		x = self._x
		dx = self._dx
		n = self._n

		numerator = 0
		for i in range(n+1):
			numerator += (-1)**(i)*math.comb(n, i)*f(x + (n/2-i)*dx)
		
		return numerator/dx**n



def richardson(f:_types.FunctionType, x:float, dx=1E-1, n=1)->float:
	"""
	Richardson Extrapolation for 1st order derivative
	Uses central differences
	"""
	h = dx
	d1 = FiniteDiff(f, x, n=n, dx=h)
	d2 = FiniteDiff(f, x, n=n, dx=h/2)

	return 4/3*d2.central() - 1/3*d1.central()




if __name__ == "__main__":
	f = lambda x: -0.1*x**4 - 0.15*x**3 - 0.5*x**2 - 0.25*x + 1.2
	f2 = lambda x: x**5
	d = FiniteDiff(f, x=0.5, n=1, dx=0.25)
	print(d.forward())
	print(d.backward())
	print(d.central())

	print(richardson(f2, x=2, n=4, dx=0.25))


import ctypes as _ct
from ._ctypeslib import pydll

pydll.c_plot_app.argtypes = []
pydll.c_plot_app.restype = None

pydll.c_plot_mainloop.argtypes = [_ct.c_bool]
pydll.c_plot_mainloop.restype=_ct.c_bool

pydll.c_plot_ismainlooprunning.argtypes = []
pydll.c_plot_ismainlooprunning.restype=_ct.c_bool

pydll.c_plot_exitmainloop.argtypes = []
pydll.c_plot_exitmainloop.restype=_ct.c_bool




class App:
	def __init__(self) -> None:
		"""
		Initiates the GUI application
		"""
		#IMPORTANT: if not set colors/text of plots looks blurry
		_ct.windll.shcore.SetProcessDpiAwareness(True)
		pydll.c_plot_app()
	

	def mainloop(self, shared=False):
		"""
		Starts the main loop \n
		shared: True, if there is any other application using a main loop \n

		returns True if mainloop is started
		"""
		return pydll.c_plot_mainloop(_ct.c_bool(shared))


	def ismainlooprunning(self):
		"""
		Is the main loop initiated by scisuit library running
		"""
		return pydll.c_plot_ismainlooprunning()


	def exitmainloop(self):
		"""
		Exits the main loop inititated by scisuit library and 
		closes all the associated windows. \n

		returns True if main loop is exited.
		"""
		return pydll.c_plot_exitmainloop()


import ctypes as _ct
from pathlib import Path as _Path
import sys



"""
Tested with: 3.10.6, 3.11.6, 3.12.0
"""
_DLLname = "scisuit_pybind" + str(sys.version_info.major) + str(sys.version_info.minor)


#__file__ is guaranteed to be an absolute path in Python 3.9+
__pt = _Path(__file__)
pydll = _ct.PyDLL(str(__pt.parents[0] / _DLLname))



__all__ = ['pydll']




"""
-------------------------------------------------------------------------------
------------------------  ENGINEERING --------------------------------------- 
"""

pydll.c_eng_psychrometry.argtypes = [_ct.py_object]
pydll.c_eng_psychrometry.restype = _ct.py_object






"""
-------------------------------------------------------------------------------
------------------------    ROOTS  --------------------------------------- 
"""

pydll.c_root_bisect.argtypes = [_ct.py_object, _ct.c_double, _ct.c_double, _ct.c_double, _ct.c_int, _ct.c_char_p, _ct.c_bool]
pydll.c_root_bisect.restype = _ct.py_object

pydll.c_root_brentq.argtypes = [_ct.py_object, _ct.c_double, _ct.c_double, _ct.c_double, _ct.c_int]
pydll.c_root_brentq.restype = _ct.py_object

pydll.c_root_muller.argtypes = [_ct.py_object, _ct.py_object, _ct.py_object, _ct.py_object, _ct.py_object, _ct.c_double, _ct.c_int]
pydll.c_root_muller.restype = _ct.py_object

pydll.c_root_newton.argtypes = [_ct.py_object, _ct.c_double, _ct.py_object, _ct.py_object, _ct.c_double, _ct.c_int]
pydll.c_root_newton.restype = _ct.py_object

pydll.c_root_ridder.argtypes = [_ct.py_object, _ct.c_double, _ct.c_double, _ct.c_double, _ct.c_int]
pydll.c_root_ridder.restype = _ct.py_object








"""
-------------------------------------------------------------------------------
------------------------  FITTING PACKAGE --------------------------------------- 
"""

pydll.c_fit_expfit.argtypes = [_ct.py_object, _ct.py_object, _ct.py_object]
pydll.c_fit_expfit.restype = _ct.py_object
pydll.c_fit_lagrange.argtypes = [_ct.py_object, _ct.py_object,_ct.c_double]
pydll.c_fit_lagrange.restype = _ct.py_object
pydll.c_fit_logfit.argtypes = [_ct.py_object, _ct.py_object]
pydll.c_fit_logfit.restype = _ct.py_object
pydll.c_fit_logistfit.argtypes = [_ct.py_object, _ct.py_object, _ct.py_object]
pydll.c_fit_logistfit.restype = _ct.py_object
pydll.c_fit_powfit.argtypes = [_ct.py_object, _ct.py_object]
pydll.c_fit_powfit.restype = _ct.py_object
pydll.c_fit_spline.argtypes = [_ct.py_object, _ct.py_object]
pydll.c_fit_spline.restype = _ct.py_object






"""
-------------------------------------------------------------------------------
------------------------  INTEGRATION --------------------------------------- 
"""
pydll.c_integ_simpson.argtypes = [_ct.py_object, _ct.py_object]
pydll.c_integ_simpson.restype = _ct.py_object

pydll.c_integ_romberg.argtypes = [_ct.py_object, _ct.c_double, _ct.c_double, _ct.c_double, _ct.c_int]
pydll.c_integ_romberg.restype = _ct.py_object

pydll.c_integ_fixed_quad.argtypes = [_ct.py_object, _ct.c_double, _ct.c_double, _ct.c_int]
pydll.c_integ_fixed_quad.restype = _ct.py_object





"""
-------------------------------------------------------------------------------
------------------------  STATS --------------------------------------- 
"""

pydll.c_stat_dbeta.argtypes = [_ct.py_object,  _ct.c_double, _ct.c_double]
pydll.c_stat_dbeta.restype=_ct.py_object
pydll.c_stat_pbeta.argtypes = [_ct.py_object,  _ct.c_double, _ct.c_double]
pydll.c_stat_pbeta.restype=_ct.py_object
pydll.c_stat_qbeta.argtypes = [_ct.py_object,  _ct.c_double, _ct.c_double]
pydll.c_stat_qbeta.restype=_ct.py_object


pydll.c_stat_dbinom.argtypes = [_ct.py_object, _ct.c_int, _ct.c_double]
pydll.c_stat_dbinom.restype=_ct.py_object
pydll.c_stat_pbinom.argtypes = [_ct.py_object, _ct.c_int, _ct.c_double]
pydll.c_stat_pbinom.restype=_ct.py_object
pydll.c_stat_qbinom.argtypes = [_ct.py_object, _ct.c_int, _ct.c_double]
pydll.c_stat_qbinom.restype=_ct.py_object


pydll.c_stat_dnbinom.argtypes = [_ct.py_object, _ct.c_int, _ct.c_double]
pydll.c_stat_dnbinom.restype=_ct.py_object
pydll.c_stat_pnbinom.argtypes = [_ct.py_object, _ct.c_int, _ct.c_double]
pydll.c_stat_pnbinom.restype=_ct.py_object
pydll.c_stat_qnbinom.argtypes = [_ct.py_object, _ct.c_int, _ct.c_double]
pydll.c_stat_qnbinom.restype=_ct.py_object


pydll.c_stat_dmultinom.argtypes = [_ct.py_object, _ct.c_int, _ct.py_object]
pydll.c_stat_dmultinom.restype=_ct.py_object


pydll.c_stat_dchisq.argtypes = [_ct.py_object, _ct.c_int]
pydll.c_stat_dchisq.restype=_ct.py_object
pydll.c_stat_pchisq.argtypes = [_ct.py_object, _ct.c_int]
pydll.c_stat_pchisq.restype=_ct.py_object
pydll.c_stat_qchisq.argtypes = [_ct.py_object, _ct.c_int]
pydll.c_stat_qchisq.restype=_ct.py_object


pydll.c_stat_dexp.argtypes = [_ct.py_object, _ct.c_double]
pydll.c_stat_dexp.restype=_ct.py_object
pydll.c_stat_pexp.argtypes = [_ct.py_object, _ct.c_double]
pydll.c_stat_pexp.restype=_ct.py_object
pydll.c_stat_qexp.argtypes = [_ct.py_object, _ct.c_double]
pydll.c_stat_qexp.restype=_ct.py_object


pydll.c_stat_df.argtypes = [_ct.py_object, _ct.c_int, _ct.c_int]
pydll.c_stat_df.restype=_ct.py_object
pydll.c_stat_pf.argtypes = [_ct.py_object, _ct.c_int, _ct.c_int]
pydll.c_stat_pf.restype=_ct.py_object
pydll.c_stat_qf.argtypes = [_ct.py_object, _ct.c_int, _ct.c_int]
pydll.c_stat_qf.restype=_ct.py_object


pydll.c_stat_dgamma.argtypes = [_ct.py_object, _ct.c_double, _ct.c_double]
pydll.c_stat_dgamma.restype=_ct.py_object
pydll.c_stat_pgamma.argtypes = [_ct.py_object, _ct.c_double, _ct.c_double]
pydll.c_stat_pgamma.restype=_ct.py_object
pydll.c_stat_qgamma.argtypes = [_ct.py_object, _ct.c_double, _ct.c_double]
pydll.c_stat_qgamma.restype=_ct.py_object


pydll.c_stat_dgeom.argtypes = [_ct.py_object, _ct.c_double]
pydll.c_stat_dgeom.restype=_ct.py_object
pydll.c_stat_pgeom.argtypes =  [_ct.py_object, _ct.c_double]
pydll.c_stat_pgeom.restype=_ct.py_object
pydll.c_stat_qgeom.argtypes =  [_ct.py_object, _ct.c_double]
pydll.c_stat_qgeom.restype=_ct.py_object


pydll.c_stat_dhyper.argtypes = [_ct.py_object, _ct.c_int, _ct.c_int, _ct.c_int]
pydll.c_stat_dhyper.restype=_ct.py_object
pydll.c_stat_phyper.argtypes = [_ct.py_object, _ct.c_int, _ct.c_int, _ct.c_int]
pydll.c_stat_phyper.restype=_ct.py_object
pydll.c_stat_qhyper.argtypes = [_ct.py_object, _ct.c_int, _ct.c_int, _ct.c_int]
pydll.c_stat_qhyper.restype=_ct.py_object


pydll.c_stat_dnorm.argtypes = [_ct.py_object, _ct.c_double, _ct.c_double]
pydll.c_stat_dnorm.restype=_ct.py_object
pydll.c_stat_pnorm.argtypes = [_ct.py_object, _ct.c_double, _ct.c_double]
pydll.c_stat_pnorm.restype=_ct.py_object
pydll.c_stat_qnorm.argtypes = [_ct.py_object, _ct.c_double, _ct.c_double]
pydll.c_stat_qnorm.restype=_ct.py_object


pydll.c_stat_dlnorm.argtypes = [_ct.py_object, _ct.c_double, _ct.c_double]
pydll.c_stat_dlnorm.restype=_ct.py_object
pydll.c_stat_plnorm.argtypes = [_ct.py_object, _ct.c_double, _ct.c_double]
pydll.c_stat_plnorm.restype=_ct.py_object
pydll.c_stat_qlnorm.argtypes = [_ct.py_object, _ct.c_double, _ct.c_double]
pydll.c_stat_qlnorm.restype=_ct.py_object


pydll.c_stat_dpareto.argtypes = [_ct.py_object, _ct.c_double, _ct.c_double]
pydll.c_stat_dpareto.restype=_ct.py_object
pydll.c_stat_ppareto.argtypes = [_ct.py_object, _ct.c_double, _ct.c_double]
pydll.c_stat_ppareto.restype=_ct.py_object
pydll.c_stat_qpareto.argtypes = [_ct.py_object, _ct.c_double, _ct.c_double]
pydll.c_stat_qpareto.restype=_ct.py_object


pydll.c_stat_dpois.argtypes = [_ct.py_object, _ct.c_double]
pydll.c_stat_dpois.restype=_ct.py_object
pydll.c_stat_ppois.argtypes = [_ct.py_object, _ct.c_double]
pydll.c_stat_ppois.restype=_ct.py_object
pydll.c_stat_qpois.argtypes = [_ct.py_object, _ct.c_double]
pydll.c_stat_qpois.restype=_ct.py_object


pydll.c_stat_dt.argtypes = [_ct.py_object, _ct.c_int]
pydll.c_stat_dt.restype=_ct.py_object
pydll.c_stat_pt.argtypes = [_ct.py_object, _ct.c_int]
pydll.c_stat_pt.restype=_ct.py_object
pydll.c_stat_qt.argtypes = [_ct.py_object, _ct.c_int]
pydll.c_stat_qt.restype=_ct.py_object


pydll.c_stat_dunif.argtypes = [_ct.py_object, _ct.c_double, _ct.c_double]
pydll.c_stat_dunif.restype=_ct.py_object
pydll.c_stat_punif.argtypes = [_ct.py_object, _ct.c_double, _ct.c_double]
pydll.c_stat_punif.restype=_ct.py_object
pydll.c_stat_qunif.argtypes = [_ct.py_object, _ct.c_double, _ct.c_double]
pydll.c_stat_qunif.restype=_ct.py_object


pydll.c_stat_dweibull.argtypes = [_ct.py_object, _ct.c_double, _ct.c_double]
pydll.c_stat_dweibull.restype=_ct.py_object
pydll.c_stat_pweibull.argtypes = [_ct.py_object, _ct.c_double, _ct.c_double]
pydll.c_stat_pweibull.restype=_ct.py_object
pydll.c_stat_qweibull.argtypes = [_ct.py_object, _ct.c_double, _ct.c_double]
pydll.c_stat_qweibull.restype=_ct.py_object


pydll.c_stat_dsignrank.argtypes = [_ct.py_object, _ct.c_int]
pydll.c_stat_dsignrank.restype=_ct.py_object
pydll.c_stat_psignrank.argtypes = [_ct.py_object, _ct.c_int]
pydll.c_stat_psignrank.restype=_ct.py_object
pydll.c_stat_qsignrank.argtypes = [_ct.py_object, _ct.c_int]
pydll.c_stat_qsignrank.restype=_ct.py_object


pydll.c_stat_moveavg.argtypes = [_ct.py_object, _ct.py_object, _ct.c_int]
pydll.c_stat_moveavg.restype=_ct.py_object

pydll.c_stat_rolling.argtypes = [_ct.py_object, _ct.py_object, _ct.c_int]
pydll.c_stat_rolling.restype=_ct.py_object

pydll.c_stat_test_norm_ad.argtypes = [_ct.py_object]
pydll.c_stat_test_norm_ad.restype=_ct.py_object






"""
--------------------------------------------------------------------------------
----------------------------  PLOT LIBRARY --------------------------------------
"""


pydll.c_plot_boxplot.argtypes = [_ct.py_object, _ct.py_object]
pydll.c_plot_boxplot.restype=_ct.py_object


pydll.c_plot_histogram.argtypes = [_ct.py_object, _ct.py_object]
pydll.c_plot_histogram.restype=_ct.py_object


pydll.c_plot_psychrometry.argtypes = [_ct.py_object, _ct.py_object]
pydll.c_plot_psychrometry.restype=_ct.py_object


pydll.c_plot_scatter.argtypes = [_ct.py_object, _ct.py_object]
pydll.c_plot_scatter.restype=_ct.py_object


pydll.c_plot_canvas.argtypes = [
						_ct.py_object, #x-bounds
						_ct.py_object, #y-bounds
						_ct.c_bool, # Show Horiz Axis
						_ct.c_bool, # Show Vert Axis
						_ct.c_bool, # Show Horiz Gridlines
						_ct.c_bool] # Show Vert Gridlines
pydll.c_plot_canvas.restype=_ct.py_object



#-----------------------------------------
pydll.c_plot_gdi_arc.argtypes = [
					_ct.c_double, #x1
					_ct.c_double, #y1
					_ct.c_double, #x2
					_ct.c_double, #y2
					_ct.c_double, #xcenter
					_ct.c_double, #ycenter
					_ct.py_object, #dictionary for Pen
					_ct.py_object] #dictionary for Brush				
pydll.c_plot_gdi_arc.restype=None


pydll.c_plot_gdi_arrow.argtypes = [
					_ct.c_double, #x1
					_ct.c_double, #y1
					_ct.c_double, #x2
					_ct.c_double, #y2
					_ct.c_double, #angle
					_ct.c_double, #length
					_ct.py_object] #dictionary for Pen				
pydll.c_plot_gdi_arrow.restype=None


pydll.c_plot_gdi_line.argtypes = [
					_ct.c_double, #x1
					_ct.c_double, #y1
					_ct.c_double, #x2
					_ct.c_double, #y2
					_ct.py_object] #dictionary for Pen				
pydll.c_plot_gdi_line.restype=None


pydll.c_plot_gdi_rect.argtypes = [
					_ct.c_double, #x of top-left
					_ct.c_double, #y of top-left
					_ct.c_double, #width
					_ct.c_double, #height
					_ct.py_object, #dictionary for Pen
					_ct.py_object] #dictionary for Brush
pydll.c_plot_gdi_ellipse.restype=None


pydll.c_plot_gdi_ellipse.argtypes = [
					_ct.c_double, # x of center
					_ct.c_double, # y of center
					_ct.c_double, # half width
					_ct.c_double, # half height
					_ct.py_object, #dictionary for Pen
					_ct.py_object] #dictionary for Brush
pydll.c_plot_gdi_ellipse.restype=None


pydll.c_plot_gdi_text.argtypes = [
					_ct.c_double, # x of top-left
					_ct.c_double, # y of top-left
					_ct.c_char_p, # text
					_ct.c_double, # rotation
					_ct.c_char_p, # color
					_ct.py_object] #dictionary for Font
pydll.c_plot_gdi_text.restype=None


pydll.c_plot_gdi_curve.argtypes = [
					_ct.py_object, #x-values (iterable)
					_ct.py_object,#y-values (iterable)
					_ct.py_object] #dictionary for Pen
pydll.c_plot_gdi_curve.restype=None


pydll.c_plot_gdi_polygon.argtypes = [
					_ct.py_object, #x-values (iterable)
					_ct.py_object,#y-values (iterable)
					_ct.py_object, #dictionary for Pen
					_ct.py_object] #dictionary for brush
pydll.c_plot_gdi_polygon.restype=None


pydll.c_plot_gdi_marker.argtypes = [
					_ct.c_double, # x of centroid
					_ct.c_double, # y of centroid
					_ct.c_char_p, # type
					_ct.c_uint8, # size
					_ct.py_object, # dictionary for Pen
					_ct.py_object] #dictionary for Brush
pydll.c_plot_gdi_marker.restype=None

#---------------------------------------------

pydll.c_plot_layout.argtypes = [_ct.c_int, _ct.c_int]
pydll.c_plot_layout.restype=None


pydll.c_plot_subplot.argtypes = [
						_ct.c_int, 
						_ct.c_int, 
						_ct.c_int, 
						_ct.c_int]
pydll.c_plot_subplot.restype=None


pydll.c_plot_figure.argtypes = []
pydll.c_plot_figure.restype=None


pydll.c_plot_title.argtypes = [_ct.c_char_p]
pydll.c_plot_title.restype=None


pydll.c_plot_xlabel.argtypes = [_ct.c_char_p]
pydll.c_plot_xlabel.restype=None


pydll.c_plot_ylabel.argtypes = [_ct.c_char_p]
pydll.c_plot_ylabel.restype=None


pydll.c_plot_axislim.argtypes = [
						_ct.py_object, 
						_ct.py_object, 
						_ct.c_char]
pydll.c_plot_axislim.restype=_ct.py_object


pydll.c_plot_set_xticks.argtypes = [
						_ct.py_object, 
						_ct.py_object,
						_ct.c_char_p, #Alignment
						_ct.c_char_p] #Position
pydll.c_plot_set_xticks.restype=_ct.py_object


pydll.c_plot_set_yticks.argtypes = [
						_ct.py_object, 
						_ct.py_object,
						_ct.c_char_p, #Alignment
						_ct.c_char_p] #Position
pydll.c_plot_set_yticks.restype=_ct.py_object

pydll.c_plot_set_axispos.argtypes = [_ct.c_double, _ct.c_char]
pydll.c_plot_set_axispos.restype=_ct.py_object

pydll.c_plot_legend.argtypes = []
pydll.c_plot_legend.restype=None


pydll.c_plot_show.argtypes = []
pydll.c_plot_show.restype=None





import numpy as np
import numbers
from typing import Iterable as _Iterable
from random import randint

from .gdi import rect
from .charts import canvas, xlim, ylim, set_xticks, set_yticks
from ..util import NiceNumbers, minmax





def bar(
		x, 
		height, 
		width=0.8, 
		bottom=0.0, 
		color = None, 
		**kwargs):
	"""
	`x:` The x coordinates of the bars or category
	`height:` The height of the bars.
	`width:` The width of the bars.
	`bottom:` The y coordinate of the bottom side of the bars.
	"""
	assert isinstance(bottom, numbers.Real|_Iterable)
	if isinstance(bottom, _Iterable):
		nums = [i for i in bottom if isinstance(i, numbers.Real)]
		assert len(nums) == len(height), "if bottom is Iterable, its length must be equal to 'height's length"
	
	pos = np.array(x)
	X_HasStr = False
	if isinstance(pos[0], str):
		pos = list(range(len(x)))
		X_HasStr = True
	
	arr = np.array(height)

	_Min = np.min(bottom) if isinstance(bottom, _Iterable) else bottom
	_Max = np.max(arr + np.array(bottom) if isinstance(bottom, _Iterable) else arr + bottom)

	PrevMin, PrevMax = ylim()
	_Min = min(_Min, PrevMin)
	_Max = max(_Max, PrevMax)

	niceY = NiceNumbers(_Min, _Max)
	canvas(
		x=[-width, len(x)], 
		y=niceY.minmax, 
		vgrid=kwargs.get("vgrid") or False,
		hgrid=kwargs.get("hgrid") if kwargs.get("hgrid")!=None else True,
		haxis=kwargs.get("haxis") if kwargs.get("haxis")!=None else True,
		vaxis=kwargs.get("vaxis") if kwargs.get("vaxis")!=None else True
		)


	_Color = color or kwargs.get("fc") or kwargs.get("facecolor")
	if _Color == None:
		_Color = [randint(0, 255), randint(0, 255), randint(0, 255)]


	for i in range(len(height)):
		_bottom = bottom if isinstance(bottom, numbers.Real) else bottom[i]
		_xcord = pos[i]-width/2 if X_HasStr else pos[i]
		xy = [_xcord, _bottom]

		kwargs["hatch"] = kwargs.get("hatch") or "solid"
		kwargs["facecolor"] = kwargs["fc"] = _Color if (isinstance(_Color, str) or isinstance(_Color[0], numbers.Real)) else _Color[i]

		rect(xy=xy, width=width, height=float(abs(height[i])), **kwargs)
	
	if X_HasStr:
		set_xticks(pos, x)





def barh(
		y, 
		width, 
		height=0.8, 
		left=0.0, 
		color = None, 
		**kwargs):
	"""
	`y:` The y coordinates of the bars or category
	`height:` The height of the bars.
	`width:` The width of the bars.
	`bottom:` The y coordinate of the bottom side of the bars.
	"""
	assert isinstance(left, numbers.Real|_Iterable)
	if isinstance(left, _Iterable):
		nums = [i for i in left if isinstance(i, numbers.Real)]
		assert len(nums) == len(width), "if left is Iterable, its length must be equal to 'width's length"
	
	pos = np.array(y)
	X_HasStr = False
	if isinstance(pos[0], str):
		pos = list(range(len(y)))
		X_HasStr = True
	
	arr = np.array(width)

	_Min = np.min(left) if isinstance(left, _Iterable) else left
	_Max = np.max(arr + np.array(left) if isinstance(left, _Iterable) else arr + left)

	PrevMin, PrevMax = xlim()
	_Min = min(_Min, PrevMin)
	_Max = max(_Max, PrevMax)

	niceX = NiceNumbers(_Min, _Max)
	canvas(
		y=[-height, len(y)], 
		x=niceX.minmax, 
		hgrid=kwargs.get("hgrid") or False,
		vgrid=kwargs.get("vgrid") if kwargs.get("vgrid")!=None else True,
		haxis=kwargs.get("haxis") if kwargs.get("haxis")!=None else True,
		vaxis=kwargs.get("vaxis") if kwargs.get("vaxis")!=None else True
		)


	_Color = color or kwargs.get("fc") or kwargs.get("facecolor")
	if _Color == None:
		_Color = [randint(0, 255), randint(0, 255), randint(0, 255)]


	for i in range(len(width)):
		_xcoord = left if isinstance(left, numbers.Real) else left[i]
		_ycoord = pos[i]-height/2 if X_HasStr else pos[i]
		xy = [_xcoord, _ycoord]

		kwargs["hatch"] = kwargs.get("hatch") or "solid"
		kwargs["facecolor"] = kwargs["fc"] = _Color if (isinstance(_Color, str) or isinstance(_Color[0], numbers.Real)) else _Color[i]

		rect(xy=xy, width=float(abs(width[i])), height=float(height), **kwargs)
	
	if X_HasStr:
		set_yticks(pos, y)


import numpy as _np
import math as _math

import ctypes as _ct
from .._ctypeslib import pydll as _pydll





def kurt(y:_np.ndarray | list):
	"""
	Computes excess kurtosis. \n
	y: ndarray / list.
	"""
	n = len(y)

	assert n >= 4, "list/ndarray must have at least 4 elements"
	assert isinstance(y, list) or isinstance(y, _np.ndarray), "list/ndarray expected"
	
	Arr = y
	if(isinstance(y, list)):
		Arr = _np.asfarray(y)

	avg = _np.mean(Arr)
	stdev = _np.std(Arr, ddof=1)

	s = 0
	for Num in y:
		s += (Num - avg)**4

	Kurt = (s/n)/(stdev)**4 - 3

	return float(Kurt)



def mode(y:_np.ndarray | list)->tuple:
	"""
	Computes mode

	## Example
	arr = np.array([1, 3, 5, 5, 7, 9, 3, 5, 7, 3, 6])
	print(mode(y = arr))
	"""
	n = len(y)

	assert n >= 3, "list/ndarray must have at least 3 elements"
	assert isinstance(y, list) or isinstance(y, _np.ndarray), "list/ndarray expected"
	
	Arr = None
	if(isinstance(y, list)):
		Arr = _np.asarray(y)
	else:
		Arr = _np.array(y)

	values, counts = _np.unique(Arr, return_counts=True)
	if(len(values) == len(Arr)):
		return ([], 0)

	index = _np.argwhere(counts == _np.max(counts))

	return values[index].flatten().tolist(), int(_np.max(counts))




def moveavg(x, y, period=2):
	return _pydll.c_stat_moveavg(x, y, _ct.c_int(period))



class rolling:

	"""
	A class to compute rolling window mean, std, var min, max...

	## Example:
	>>r = rolling(x, y, period=2) \n
	>>r.mean() \n
	>>r.var() \n \n

	OR \n
	
	>>r=rolling(x, y, period=2).mean() \n

	If multiple calls required, prefer the first way for efficiency.

	"""

     
	def __init__(self, x:list, y:list, period:int = 2):
		assert isinstance(x, list) or isinstance(x, _np.ndarray), "x must be of type list/ndarray"
		assert isinstance(y, list) or isinstance(y, _np.ndarray), "y must be of type list/ndarray"
		assert isinstance(period, int), "period must be of type int"

		self._m_X, self._m_Windows = _pydll.c_stat_rolling(x=x, y=y, period=period)
		

	def __compute(self, func)->list:
		"""
		Calls the parameter func on self._m_Windows
		Returns a 1D list containing real numbers
		"""
		retList = []
		for Lst in self._m_Windows:
			retList.append(func(Lst))
		
		return retList
	

	def get(self):
		"""
		When mean, median, min, max member funcs do not meet the purpose.\n
		To process rolling windows data (2D list) manually. n
		get()->(list, 2D list)
		"""
		return self._m_X, self._m_Windows
	

	def mean(self):
		return self.__compute(self.__mean)

	def median(self):
		return self.__compute(self.__median)
	
	def min(self):
		return self.__compute(min)
	
	def max(self):
		return self.__compute(max)

	

	def __mean(self, lst:list):
		return sum(lst)/len(lst)

	def __median(self, arg):
		arr:_np.ndarray = arg

		if isinstance(arg, list):
			arr = _np.asfarray(arg)
		
		return float(_np.median(arr))




def skew(y:_np.ndarray | list):
	"""
	Computes the skewness of a distribution

	## Example
	y = _np.random.standard_normal(1000)
	print(skew(y))
	
	"""
	n = len(y)

	assert n >= 3, "list/ndarray must have at least 3 elements"
	assert isinstance(y, list) or isinstance(y, _np.ndarray), "list/ndarray expected"
	
	Arr = None
	if(isinstance(y, list)):
		Arr = _np.asarray(y)
	else:
		Arr = _np.array(y)
	
	avg, std_p = _np.mean(Arr), _np.std(Arr)

	skew_p = _np.sum((Arr-avg)**3)
	skew_p /= (n * std_p**3.0);

	return _math.sqrt(n * (n - 1)) * skew_p/(n-2)




def sample(y, size:int, replace = False):
	"""
	y: ndarray / list.

	## Example
	ss = _np.linspace(start=1, stop=10, num=10) \n
	s = sample(ss, 4) \n
	s1 = sample(ss, 11, replace=True) \n
	s2 = sample(ss, 11, replace=False) \n
	print(s1)
	"""

	return _np.random.choice(a=y, size=size,  replace=replace)


"""
BASIC STATISTICAL TESTS:

* Anderson-Darling Normality Test
* F-test
* Sign-Test
* t-tests (1-sample, 2-sample, paired)
* Z-test
"""


import math
import numbers
from typing import Iterable
import numpy as _np
from .._ctypeslib import pydll as _pydll

from dataclasses import dataclass
from .distributions import pbinom, pf, qf, pt, qt, pnorm, qnorm



__all__ = ['test_norm_ad', 'TestNormRes',
		'test_f', 'test_f_Result', 
		'test_sign', 'test_sign_Result', 'CI_Result',
		'test_t', 'test_t1_result', 'test_t2_result', 'test_tpaired_result', 
		'test_z', 'test_z_Result' ]







""" *********** Anderson-Darling Test ******************* """

@dataclass
class TestNormRes:
	pval:float
	A2:float

def test_norm_ad(x:Iterable)->TestNormRes:
	assert isinstance(x, Iterable), "x must be an Iterable object"
	
	pval, A2 = _pydll.c_stat_test_norm_ad(x)
	return TestNormRes(pval, A2)









""" *********** F-Test ******************* """

@dataclass
class test_f_Result:
	Fcritical:float
	df1:int; df2:int
	var1:float; var2:float
	CI_lower:float; CI_upper:float



def test_f(
		x:Iterable, 
		y:Iterable, 
		ratio:float = 1.0, 
		alternative:str = "two.sided", 
		conflevel:float = 0.95)->tuple[float, test_f_Result]:
	"""
	Performs F test

	## Return
	p-value and test_f_Result class.

	## Input
	x/y: First/second sample
	alternative: "two.sided", "less", "greater"
	ratio: Assumed ratio of variances of the samples
	conflevel: Confidence level, [0,1] 
	"""
	assert conflevel>=0.0 or conflevel <= 1.0, "conflevel must be in range (0, 1)"
	assert isinstance(x, list) or type(x)==_np.ndarray, "x must be list/ndarray"
	assert isinstance(y, list) or type(y)==_np.ndarray, "y must be list/ndarray"

	XX, YY = x, y
	if isinstance(x, list):
		XX = _np.asfarray(x)

	if isinstance(y, list):
		YY = _np.asfarray(y)
	
	assert _np.issubdtype(XX.dtype, _np.number), "x must contain only numbers"
	assert _np.issubdtype(YY.dtype, _np.number), "y must contain only numbers"

	alpha = 1 - conflevel
	df1, df2 = len(XX) - 1, len(YY) -1 #degrees of freedoms
	
	var1, var2 = _np.var(XX, ddof = 1), _np.var(YY, ddof = 1)
	varRatio = float(var1) / float(var2)

	Fcritical = varRatio / ratio

	pvalue = 0.0
	if alternative == "two.sided" or alternative == "notequal":
		#F distribution is non - symmetric
		a = pf(Fcritical, df1, df2)
		b = 1 - pf(Fcritical, df1, df2)

		c = pf(1 / Fcritical, df1, df2)
		d = 1 - pf(1 / Fcritical, df1, df2)

		pvalue = min(a, b) + min(c, d)

	elif alternative == "greater":
		pvalue = (1 - pf(Fcritical, df1, df2)) #area on the right

	elif alternative == "less":
		pvalue = pf(Fcritical, df1, df2) #area on the left

	else:
		raise ValueError("Values for 'alternative': \"two.sided\" or \"notequal\", \"greater\", \"less\"");


	CI_lower, CI_upper = None, None
	if alternative == "two.sided" or alternative == "notequal":
		CI1 = varRatio * qf(alpha / 2.0, df1, df2)
		CI2 = varRatio * qf(1 - alpha / 2.0, df1, df2)

		CI_lower = min(CI1, CI2)
		CI_upper = max(CI1, CI2)

	elif alternative == "greater":
		CI_lower = varRatio * qf(alpha, df1, df2)

	elif alternative == "less":
		CI_upper = varRatio * qf(1 - alpha, df1, df2)
	
	result = test_f_Result(Fcritical= Fcritical, 
			df1=df1,
			df2=df2,
			var1=var1,
			var2=var2,
			CI_lower=CI_lower,
			CI_upper=CI_upper)

	return pvalue, result







""" ****************  Sign-Test *********************** """

@dataclass
class CI_Result:
	pos:int
	prob:float
	CILow:float; CIHigh:float


@dataclass
class test_sign_Result:
	lower:CI_Result
	interpolated:CI_Result
	upper:CI_Result


def _FindCI(xvec, alternative, ConfLevel):
	xvec = _np.sort(xvec)

	lower, exact, upper = 0.0, 0.0, 0.0
	lower_found, upper_found = False, False
	lower_pos, upper_ps = 0, 0
	prob = 0.0

	NSample = len(xvec)

	for i in range(NSample):
		_pbinom = pbinom(i, NSample, 0.5)

		if alternative == "two.sided" or alternative == "notequal":
			prob = 1 - 2 * _pbinom

		elif alternative == "greater":
			prob = 1 - _pbinom

		elif alternative == "less":
			prob = _pbinom


		if prob >= ConfLevel:
			upper = prob
			upper_found = True

			upper_pos = i

			if (lower_found): 
				break

		else:
			lower = prob
			lower_found = True

			lower_pos = i

			if (upper_found): 
				break

	CI_a = xvec[lower_pos]
	CI_b = xvec[len(xvec) - lower_pos - 1]

	lowerRes = CI_Result(
		pos = lower_pos,
		prob=lower,
		CILow = min(CI_a, CI_b),
		CIHigh = max(CI_a, CI_b))
	

	CI_a = xvec[upper_pos]
	CI_b = xvec[len(xvec) - upper_pos - 1]

	upperRes = CI_Result(
		pos = upper_pos,
		prob = upper,
		CILow = min(CI_a, CI_b),
		CIHigh = max(CI_a, CI_b))

	return lowerRes, upperRes





def test_sign(
		x:Iterable, 
		md:numbers.Real, 
		y=None, 
		alternative="two.sided", 
		conflevel=0.95)->tuple[float, test_sign_Result]:
	"""
	returns p-value and a test_sign_Result class. \n

	x: Sample
	y: Second sample
	md: Median of the population tested by the null hypothesis
	alternative: "two.sided", "less", "greater" 
	conflevel:	Confidence level, [0,1]
	"""

	NORMALAPPROX = 12

	assert conflevel>0.0 or conflevel<1.0, "conflevel must be in range (0, 1)"
	assert isinstance(x, Iterable), "x must be Iterable"

	XX = _np.asfarray(x)
	
	assert _np.issubdtype(XX.dtype, _np.number), "x must contain only numbers"
	
	xvec = XX

	if y != None:
		assert isinstance(y, Iterable), "y must be Iterable"
		YY = _np.asfarray(y)

		assert _np.issubdtype(YY.dtype, _np.number), "y must contain only numbers"
		xvec = XX - YY
	
	NGreater = len(_np.argwhere(xvec>md))

	EqIndex = _np.where(xvec == md)
	xvec = _np.delete(xvec, EqIndex)

	pvalue = 0.0

	NSample = len(xvec)

	if NSample < NORMALAPPROX:
		if alternative == "two.sided" or alternative == "notequal":
			n = min(NGreater, NSample - NGreater)

			#B(n, 0.5) is symmetric
			pvalue = 2.0 * pbinom(n, NSample, 0.5)

		elif alternative == "greater":
			pvalue = 1.0 - pbinom(NGreater - 1, NSample, 0.5) #area on the right

		elif alternative == "less":
			pvalue = pbinom(NGreater, NSample, 0.5); #area on the left
	
	else:
		variance = NSample / 4.0
		zvalue = (NGreater - NSample / 2.0) / math.sqrt(variance)

		if alternative == "two.sided" or alternative == "notequal":
			if zvalue <= 0:
				#area on the left + area on the right of positive
				pvalue = pnorm(zvalue, 0.0, 1.0) + (1 - pnorm(-1.0 * zvalue, 0.0, 1.0))
			else:
				#area on the right of positive + area on the left of negative
				pvalue = (1.0 - pnorm(zvalue, 0.0, 1.0)) + pnorm(-zvalue, 0.0, 1.0)
		
		elif alternative == "greater":
			#area on the right
			pvalue = (1.0 - pnorm(zvalue, 0.0, 1.0))

		elif alternative == "less":
			#area on the left
			pvalue = pnorm(zvalue, 0.0, 1.0)

	Lower, Upper = _FindCI(xvec, alternative, conflevel)

	interped = CI_Result(pos=-1, prob=conflevel, CILow=0.0, CIHigh=0.0)
	if alternative == "two.sided" or alternative == "notequal":
		x1, x2 = Lower.prob, Upper.prob

		interped.CILow = _np.interp(conflevel, [x1, x2], [Lower.CILow, Upper.CILow] )
		interped.CIHigh = _np.interp(conflevel, [x1, x2], [Lower.CIHigh, Upper.CIHigh])
	
	elif alternative == "greater":
		interped.CILow = _np.interp(conflevel, [Lower.prob, Upper.prob], [Lower.CILow, Upper.CILow])
		interped.CIHigh = math.inf
	
	elif alternative == "less":
		interped.CILow = -math.inf
		interped.CIHigh = _np.interp(conflevel, [Lower.prob, Upper.prob], [Lower.CIHigh, Upper.CIHigh])
	
	else:
		raise ValueError("Values for 'alternative': \"two.sided\" or \"notequal\", \"greater\", \"less\"")

	result = test_sign_Result(
		lower= Lower,
		interpolated=interped,
		upper = Upper)
	
	return pvalue, result









""" **************   t- Test ******************* """

@dataclass
class test_t1_result:
	CI_lower:float; CI_upper:float
	SE:float
	N:int
	stdev:float
	mean: float
	tcritical:float


@dataclass
class test_t2_result:
	CI_lower:float; CI_upper:float
	tcritical:float
	n1:int; n2:int
	df:int
	xaver:float; yaver:float
	s1:float; s2:float; sp:float


@dataclass
class test_tpaired_result:
	CI_lower:float; CI_upper:float
	tcritical:float
	xaver:float; yaver:float
	s1:float; s2:float; SE:float
	N:int
	mean:float #mean of difference
	stdev:float #stdev of difference


def _test_t1(
		x:Iterable, 
		mu:numbers.Real, 
		alternative="two.sided", 
		conflevel=0.95)->tuple[float, test_t1_result]:
	assert conflevel>=0.0 or conflevel <= 1.0, "conflevel must be in range (0, 1)"
	assert isinstance(x, Iterable), "x must be Iterable"

	XX = _np.asfarray(x)

	assert len(XX)>= 3, "container must have at least 3 elements"
	assert _np.issubdtype(XX.dtype, _np.number), "x must contain only numbers"

	N = len(XX)
	df = N -1
	
	stdev = _np.std(XX, ddof =1) #sample's standard deviation
	SE = stdev / math.sqrt(N) #Standard Error of Mean
	
	xaver = _np.mean(XX)
	tcritical = float((xaver - mu) / SE)

	pvalue = 0.0
	if alternative == "two.sided" or alternative == "notequal":
		if (tcritical <= 0.0):
			#area on the left of tcrit + area on the right of positive
			pvalue = pt(tcritical, df) + (1.0 - pt(abs(tcritical), df))
		else:
			#area on the right of positive tcritical + area on the left of negative tcritical
			pvalue = (1.0 - pt(tcritical, df)) + pt(-tcritical, df)
	
	#area on the right
	elif alternative == "greater":
		pvalue = (1.0 - pt(tcritical, df))
	
	#area on the left
	elif alternative == "less":
		pvalue = pt(tcritical, df); 

	else:
		raise ValueError("Values for 'alternative': \"two.sided\" or \"notequal\", \"greater\", \"less\"")
	
	alpha = 1.0 - conflevel
	CI_upper = xaver - qt(alpha / 2.0, df) * SE
	CI_lower = xaver + qt(alpha / 2.0, df) * SE

	Result = test_t1_result(
		CI_lower=CI_lower,
		CI_upper=CI_upper,
		SE = SE,
		N = N,
		stdev = stdev,
		mean = xaver,
		tcritical = tcritical)
	
	return pvalue, Result



def _test_t2(
		x:Iterable, 
		y:Iterable, 
		mu:numbers.Real, 
		varequal = True, 
		alternative="two.sided", 
		conflevel=0.95)->tuple[float, test_t2_result]:
	assert conflevel>=0.0 or conflevel <= 1.0, "conflevel must be in range (0, 1)"
	assert isinstance(x, Iterable), "x must be Iterable"

	XX = _np.asfarray(x)
	YY = _np.asfarray(y)

	assert len(XX)>= 3, "x must have at least 3 elements"
	assert len(YY)>= 3, "y must have at least 3 elements"
	assert _np.issubdtype(XX.dtype, _np.number), "x must contain only numbers"
	assert _np.issubdtype(YY.dtype, _np.number), "y must contain only numbers"

	n1, n2 = len(XX), len(YY)
	xaver, yaver = float(_np.mean(XX)), float(_np.mean(YY))
	s1, s2 = float(_np.std(XX, ddof = 1)), float(_np.std(YY, ddof = 1))
	var1, var2 = s1**2, s2**2
	alpha = 1 - conflevel

	tcritical, SE = None, None
	sp = -1 #pooled

	if varequal == False:
		df_num = (var1 / n1 + var2 / n2)**2
		df_denom = 1 / (n1 - 1) * (var1 / n1)**2 + 1 / (n2 - 1) * (var2 / n2)**2
		df = math.floor(df_num / df_denom)

		tcritical = float((xaver - yaver) - mu) / math.sqrt(var1 / n1 + var2 / n2)
		SE = math.sqrt(var1 / n1 + var2 / n2)
	
	else:
		df = n1 + n2 - 2

		sp_num = (n1 - 1) * var1 + (n2 - 1) * var2
		sp = math.sqrt(sp_num / df)

		tcritical = float((xaver - yaver) - mu) / (sp * math.sqrt(1 / n1 + 1 / n2))
		SE = sp * math.sqrt(1 / n1 + 1 / n2)
	
	pvalue = 0
	if alternative == "two.sided" or alternative == "notequal":
		if tcritical <= 0:
			#area on the left of tcrit + area on the right of positive
			pvalue = pt(tcritical, df) + (1 - pt(abs(tcritical), df))
		else:
			#area on the right of positive tcritical + area on the left of negative tcritical
			pvalue = (1 - pt(tcritical, df)) + pt(-tcritical, df)
	
	# area on the right
	elif alternative == "greater":	
		pvalue = 1.0 - pt(tcritical, df)
	
	# area on the left
	elif alternative == "less":	
		pvalue = pt(tcritical, df)
	
	else:
		raise ValueError("Values for 'alternative': \"two.sided\" or \"notequal\", \"greater\", \"less\"")
	
	quantile = qt(alpha / 2, df)

	CI1 = (xaver - yaver) - quantile * SE
	CI2 = (xaver - yaver) + quantile * SE

	CI_lower = min(CI1, CI2)
	CI_upper = max(CI1, CI2)

	Result = test_t2_result(
		CI_lower=CI_lower, CI_upper=CI_upper,
		tcritical=tcritical,
		n1 = n1, n2= n2,
		df=df,
		s1 = s1, s2=s2, sp=sp,
		xaver = xaver, yaver = yaver)
	
	return pvalue, Result


def _test_t_paired(x, y, mu, alternative="two.sided", conflevel=0.95)->tuple[float, test_tpaired_result]:
	assert conflevel>=0.0 or conflevel <= 1.0, "conflevel must be in range (0, 1)"
	assert isinstance(x, Iterable), "x must be Iterable"

	XX = x
	if isinstance(x, list):
		XX = _np.asfarray(x)
	
	YY = y
	if isinstance(y, list):
		YY = _np.asfarray(y)

	assert len(XX)>= 3, "container must have at least 3 elements"
	assert _np.issubdtype(XX.dtype, _np.number), "x must contain only numbers"
	assert _np.issubdtype(YY.dtype, _np.number), "y must contain only numbers"
	assert len(XX) == len(YY), "x and y must have same size"

	xaver, yaver = _np.mean(XX), _np.mean(YY)
	s1, s2 = _np.std(XX, ddof = 1), _np.std(YY, ddof = 1)
	
	Diff = XX- YY
	pvalue, Res1 = _test_t1(x =Diff, mu = mu, alternative = alternative, conflevel = conflevel)

	Result = test_tpaired_result(
		CI_lower=Res1.CI_lower,
		CI_upper = Res1.CI_upper,
		tcritical= Res1.tcritical,
		xaver=xaver,
		yaver=yaver,
		s1=s1,
		s2= s2,
		SE = Res1.SE,
		mean=Res1.mean,
		N = Res1.N,
		stdev=Res1.stdev)
	
	return pvalue, Result



def test_t (
		x:Iterable, 
		y:Iterable|None = None, 
		varequal=True, 
		alternative="two.sided", 
		mu:numbers.Real=0.0, 
		conflevel:numbers.Real=0.95, 
		paired=False ):
	"""
	Performs paired, 1-sample and 2-sample t-test

	x, y: First and second samples
	varequal: assuming equal variances
	alternative: 'two.sided', 'less' or 'greater'
	mu: Assumed difference between samples or assumed mean
	conflevel: Confidence level, [0,1]
	paired: For paired t-test
	"""
	if y == None:
		return _test_t1(x=x, mu=mu, alternative=alternative, conflevel=conflevel )	
	else:
		if paired==False:
			return _test_t2(x=x, y=y, mu=mu, varequal=varequal, alternative=alternative, conflevel=conflevel)
		else:
			return _test_t_paired(x=x, y=y, mu=mu, alternative=alternative, conflevel=conflevel)









""" ************* Z-Test ****************** """

@dataclass
class test_z_Result:
	SE:float ; stdev:float 
	mean: float 
	zcritical:float 
	CI_upper: float ; CI_lower: float
	N:int


def test_z(
		x:Iterable, 
		sd:numbers.Real, 
		mu:numbers.Real, 
		alternative="two.sided", 
		conflevel=0.95)->tuple[float, test_z_Result]:
	"""
	
	## Return
	p-value and test_z_Result class.

	## Input
	x: sample,  ndarray/list 
	sd: Standard deviation of population
	mu: Assumed mean of population
	alternative: "two.sided", "less", "greater"
	conflevel: Confidence level, (0,1)
	"""
	assert conflevel>0.0 or conflevel<1.0, "conflevel must be in range (0, 1)"
	assert sd >= 0, "sd must be >0"
	assert isinstance(x, Iterable), "x must be Iterable"

	XX = x
	if isinstance(x, list):
		XX = _np.asfarray(x)

	assert len(XX)>= 3, "container must have at least 3 elements"
	assert _np.issubdtype(XX.dtype, _np.number), "x must contain only numbers"

	dim = len(XX)
	xaver = _np.mean(XX)
	SE = sd / math.sqrt(dim)
	stdeviation = _np.std(XX, ddof =1); #sample's standard deviation

	alpha = 1 - conflevel
	zcritical = float((xaver - mu) / SE)

	pvalue = 0.0

	if alternative == "two.sided" or alternative == "notequal":
		if zcritical <= 0:
			#area on the left of zcrit + area on the right of positive
			pvalue = pnorm(zcritical, 0.0, 1.0) + (1.0 - pnorm(abs(zcritical), 0.0, 1.0))

		elif zcritical > 0:
			#area on the right of positive zcriticial + area on the left of negative zcriticial
			pvalue = (1.0 - pnorm(zcritical, 0.0, 1.0)) + pnorm(-zcritical, 0.0, 1.0)

	elif alternative == "greater":
		pvalue = (1.0 - pnorm(zcritical, 0.0, 1.0)) #area on the right

	elif alternative == "less":
		pvalue = pnorm(zcritical, 0.0, 1.0) #area on the left

	else:
		raise ValueError("'alternative': \"two.sided\" or \"notequal\", \"greater\", \"less\"")
	
	
	CI_upper = xaver - qnorm(alpha / 2.0, 0.0, 1.0) * SE
	CI_lower = xaver + qnorm(alpha / 2.0, 0.0, 1.0) * SE

	result = test_z_Result(SE = SE, 
			stdev = stdeviation, 
			N = dim, 
			mean = xaver, 
			zcritical = zcritical,
			CI_lower = CI_lower,
			CI_upper = CI_upper)

	return pvalue, result


from ._ctypeslib import pydll as _pydll
import ctypes as _ct
import numbers as _numbers
import types as _types
import numpy as _np
import dataclasses as _dc
import sys as _sys


__all__ = ['bisect', 'brentq', 'muller', 'newton', 'ridder', 'fsolve', "Info"]


@_dc.dataclass
class Info:
	"""
	err: error (if available)
	iter: number of iterations to reach the root
	conv: whether converged to a root or not
	msg: if convergence is False, a reason is given
	"""
	err:float = None
	iter:int = -1
	conv:bool = False
	msg:str =""



def bisect(
	f:_types.FunctionType, 
	a:float, 
	b:float, 
	tol=1E-5, 
	maxiter=100, 
	method="bf", 
	modified=False)->tuple[float, Info]:
	"""
	Finds the root using bisection method, returns (root, Info)

	## Inputs:
	f: A unary function
	a, b: The interval where the root lies in
	tol: tolerance for error
	maxiter: Maximum number of iterations
	method: "bf" for brute-force (halving) 
	"rf" for regula falsi (false position) 
	modified: True for modified regula falsi method.
	"""
	assert callable(f), "f must be function"
	assert isinstance(a, _numbers.Real), "a must be real number"
	assert isinstance(b, _numbers.Real), "b must be real number"

	root, lst =_pydll.c_root_bisect(f, _ct.c_double(a), _ct.c_double(b), 
			_ct.c_double(tol), 
			_ct.c_int(maxiter), 
			_ct.c_char_p(method.encode('utf-8')),
			_ct.c_bool(modified))
	
	return root, Info(lst[0], lst[1], lst[2], lst[3])




def brentq(
	f:_types.FunctionType, 
	a:float, 
	b:float, 
	tol=1E-5, 
	maxiter=100)->tuple[float, Info]:
	"""
	Uses the Brent's method (1973) to find the root of the function 
	using inverse quadratic interpolation, returns (root, Info)

	## Inputs:
	f: A unary function whose root is sought after \n
	a, b: The interval where root lies in \n
	tol: tolerance for error \n
	maxiter: Maximum number of iterations during the search for the root
	"""
	assert callable(f), "f must be function"
	assert isinstance(a, _numbers.Real), "a must be real number"
	assert isinstance(b, _numbers.Real), "b must be real number"

	root, lst = _pydll.c_root_brentq(f, _ct.c_double(a), _ct.c_double(b), _ct.c_double(tol), _ct.c_int(maxiter))

	return root, Info(None, lst[0], lst[1], lst[2])




def muller(
	f:_types.FunctionType, 
	x0:_numbers.Complex, 
	h=None, 
	x1=None, 
	x2=None, 
	tol=1E-5, 
	maxiter=100)->tuple[float, Info]:
	"""
	Finds root of an equation using Muller method, returns (root, Info)
	## Inputs:
	f: A unary function \n
	x0, x1, x2:	Initial guesses \n
	h: Step length \n
	tol: tolerance for error \n
	maxiter: Max number of iterations
	"""
	assert callable(f), "f must be function"
	assert isinstance(x0, _numbers.Complex), "x0 must be a Complex/Real number"
	
	root, lst = _pydll.c_root_muller(f, x0, h, x1, x2, _ct.c_double(tol), _ct.c_int(maxiter))

	return root, Info(None, lst[0], lst[1], lst[2])



def newton(
	f:_types.FunctionType, 
	x0:float, 
	x1=None, 
	fprime=None, 
	tol=1E-5, 
	maxiter=100)->tuple[float, Info]:
	"""
	If fprime is provided then uses Newton-Raphson method  \n
	If fprime is not provided, then x1 must be provided uses Secant method.

	returns (root, Info)

	## Inputs:
	f: A unary function 
	fprime: derivative of f 
	x0, x1: Initial guesses 
	tol: tolerance for error 
	maxiter: Max number of iterations
	"""
	assert callable(f), "f must be function"
	assert isinstance(x0, _numbers.Real)
	if(fprime == None):
		assert isinstance(x1, _numbers.Real), "If fprime not provided, x1 must be a real number"
	else:
		assert callable(fprime), "If fprime is provided, it must be of type function."

	root, lst = _pydll.c_root_newton(f, _ct.c_double(x0), x1, fprime, _ct.c_double(tol), _ct.c_int(maxiter))

	return root, Info(lst[0], lst[1], lst[2], lst[3])




def ridder(
	f:_types.FunctionType, 
	a:float, 
	b:float, 
	tol=1E-5, 
	maxiter=100)->tuple[float, Info]:
	"""
	Uses Ridder's method.

	## Inputs:
	f: A unary function 
	a, b: The interval where the root lies in 
	tol: tolerance for error 
	maxiter: Maximum number of iterations
	"""
	assert callable(f), "f must be function"
	assert isinstance(a, _numbers.Real), "a must be real number"
	assert isinstance(b, _numbers.Real), "b must be real number"
	
	root, lst = _pydll.c_root_ridder(f, _ct.c_double(a), _ct.c_double(b), _ct.c_double(tol), _ct.c_int(maxiter))

	return root, Info(None, lst[0], lst[1], lst[2])




def fsolve(
		F:list[_types.FunctionType], 
		x0:list[float], 
		tol=1E-5, 
		maxiter=100 )->tuple:
	"""
	Solves a system of non-linear equations using Newton's approach. \n
	Functions are in the format of f(x1,x2,...)=0
	

	## Inputs:
	F: a list of functions \n
	x0: a list of initial guesses \n


	## EXAMPLE
	x^2 + y^2 = 5 \n
	x^2 - y^2 = 1 \n

	First define the functions, F(x, y) = 0: \n

	def f1(t): \n
		return t[0]**2 + t[1]**2 - 5 \n

	def f2(t): \n
		return t[0]**2 - t[1]**2 - 1 \n
	\n
	roots, iter=fsolve( [f1,f2], [1,1] ) \n
	\n
	print(roots, "  iter:", iter) \n
	1.73205	1.41421	iter:5 \n

	print(f1(roots), " ", f2(roots)) \n
	9.428e-09    9.377e-09 
	
	"""
	assert isinstance(F, list), "F must be a list of functions"
	assert isinstance(x0, list), "x0 must be a list of numbers"

	assert isinstance(tol, _numbers.Number) and tol>0, "tol>0 expected."
	assert isinstance(maxiter, int) and maxiter>0, "maxiter>0 expected."

	dim = len(F)

	assert dim>=2, "At least 2 functions are required"
	assert dim == len(x0), "F and x0 must have same length"


	#solution vector as floating point
	v = _np.asfarray(x0)
      
	#values of each function	
	Fvals = _np.zeros(dim)

	Jacobi = _np.zeros((dim, dim)) 

	for iter in range(maxiter):
		maxfuncval = 0 #convergence criteria
	
		for i in range(dim):
			func = F[i]     #function

			assert isinstance(func, _types.FunctionType), "Entries of F must be functions of form f(t) = 0"
			
			Fvals[i] = func(v.tolist())
		
			for j in range(dim):
				oldval = v[j]
				
				#Note that vector contains (xi+dx,...)
				v[j] += tol  
				
				#evaluate function with (xi+dx,...)
				f_dxi = func(v.tolist()) 
				
				#restore the old value, vector again contains (xi,...)
				v[j] = oldval
				
				#evaluate function with (xi,...)
				f_xi = func(v.tolist())  
				
				if(abs(maxfuncval) < abs(f_xi)): 
					maxfuncval = abs(f_xi) 
				
				#register the derivative with respect to xi to Jacobian matrix
				Jacobi[i, j] = (f_dxi - f_xi) / tol


		#return solution vector and number of iterations
		if(abs(maxfuncval) < tol): 
			return v.tolist(),  iter

		DetJacobi = abs(_np.linalg.det(Jacobi))

		if(DetJacobi <= tol):
			raise RuntimeError("At iter="+ str(iter) + " Jacobian Det=" + str(DetJacobi) + ", try different initial values") 
			
		v = v - _np.linalg.solve(Jacobi, Fvals)






from .._ctypeslib import pydll as _pydll
from numbers import Real as _Real


class PsychrometryResult:
	"""
	Computation results of psychrometry class.

	## Example:
	>>result = psy.compute() \n
	>>result.P #access to pressure
	"""

	def __init__(self, psyresult):
		self._Results = psyresult

	def __str__(self) -> str:

		if(len(self._Results) == 0):
			return ""
		
		s=""
		s += "P=" + str(round(self.P, 2)) + " kPa" + "\n"
		s += "Tdb=" + str(round(self.Tdb, 2)) + " C" + "\n"	
		s += "Twb=" + str(round(self.Twb, 2)) + " C" + "\n"	
		s += "Tdp=" + str(round(self.Tdp, 2)) +  " C" + "\n"	
		s += "H=" + str(round(self.H, 2)) + " kJ/kg da" + "\n"	
		s += "RH=" + str(round(self.RH, 2)) + " %" + "\n"	
		s += "W=" + str(round(self.W, 4)) +  " kg/kg da" + "\n"	
		s += "V=" + str(round(self.V,3)) +  " m3/kg da" + "\n"

		return s


	@property
	def P(self): 
		"""Pressure in kPa"""
		return self._Results.get("P")

	@property
	def Tdb(self): 
		"""Dry bulb temperature in Celcius"""
		return self._Results.get("Tdb")

	@property
	def Twb(self): 
		"""Wet bulb temperature in Celcius"""
		return self._Results.get("Twb")

	@property
	def Tdp(self): 
		"""Dew point temperature in Celcius"""
		return self._Results.get("Tdp")

	@property
	def H(self):
		"""Enthalpy in kJ/kg da""" 
		return self._Results.get("H")

	@property
	def RH(self): 
		"""Relative Humidity in %"""
		return self._Results.get("RH")

	@property
	def W(self):
		"""Absolute Humidity in kg/kg da""" 
		return self._Results.get("W")

	@property
	def V(self): 
		"""Specific volume in m3/kg da"""
		return self._Results.get("V")

	@property
	def Pws(self):
		"""water vapor pressure at saturation in kPa"""
		return self._Results.get("Pws")

	@property
	def Pw(self):
		"""water vapor pressure in kPa"""
		return self._Results.get("Pw")

	@property
	def Ws(self):
		"""Absolute humidity at saturation kg/kg da"""
		return self._Results.get("Ws")



def psychrometry(
		P:_Real=None,
		Tdb:_Real=None, 
		Twb:_Real=None, 
		Tdp:_Real=None, 
		W:_Real=None, 
		RH:_Real=None, 
		H:_Real=None, 
		V:_Real=None):
		
	"""
	Computes thermodynamic properties of humid-air.\n 
	Only 3 parameters must have Real values

	## Input: 
	P: Pressure (kPa)
	Tdb, Twb, Tdp: Dry-bulb, wet-bulb and dew point temperatures (°C)
	W, RH: Absolute(kg/kg da) and relative humidity (%)
	H: Enthalpy (kJ/kg da)
	V: Specific Volume (m3/ kg da)

	## Example
	p=psychrometry(P=100, Tdb=50, RH=60) \n
	print(p) # prints all properties \n
	p.H # access only to enthalpy

	"""
	d = {"P": P, "Tdb":Tdb, "Twb":Twb, "Tdp":Tdp, "W":W, "RH":RH, "H":H, "V":V}

	filtered = {k:v for k, v in d.items() if isinstance(v, _Real)}
	assert len(filtered)==3, "Exactly 3 parameters with real values expected."

	return PsychrometryResult(_pydll.c_eng_psychrometry(filtered))



import numbers

class Pen:
	def __init__(self, *args):

		assert isinstance(args[0], dict), "first argument must be dict"

		params:dict = args[0]

		self.color = params.get("edgecolor") or params.get("ec")
		if self.color != None:
			assert isinstance(self.color, str|tuple|list), "'edgecolor' must be str|tuple|list"
		
		self.alpha = params.get("alpha")
		assert isinstance(self.alpha, None|numbers.Real)

		self.style = params.get("linestyle") or params.get("ls")
		if self.style == None:
			self.style = "-"
		else:
			assert isinstance(self.style, str), "'style' must be str"
		
		_style1 = ["-",":","---", "--", "-.", "" ]
		_style2 = ["solid","dotted","ldashed", "dashed", "dashdot", "none" ]

		index = _style1.index(self.style) if self.style in _style1 else -1
		if index<0:
			index2 = _style2.index(self.style) if self.style in _style2 else -1

			#user has not provided from either style1 or style2, raise error
			if index2<0:
				raise ValueError(",".join(_style1) + " or " + ",".join(_style2))
			else:
				#located from style2, get equivalent style1
				self.style = _style1[index2]


		self.width = params.get("linewidth") or params.get("lw")
		if self.width == None:
			self.width = 1
		else:
			assert isinstance(self.width, int), "'width' must be integer"
			assert self.width>0, "width>0 expected"
	
	def __iter__(self):
		return iter([
			("style",self.style),
			("width", self.width),
			("color", self.color),
			("alpha", self.alpha)])	






__SOLID = 100 #"solid"
__TRANSPARENT = 106 #"none"
__BDIAGHATCH = 111 # "\"
__CROSSDIAGHATCH = 112 #"x"
__FDIAGHATCH = 113 #"/"
__CROSSHATCH = 114 #"+"
__HORIZHATCH =115 #"-"
__VERTHATCH = 116 # |

	
class Brush:
	def __init__(self, *args):
		assert isinstance(args[0], dict), "first argument must be dict"
		params:dict = args[0]

		self.color = params.get("facecolor") or params.get("fc")
		if self.color != None:
			assert isinstance(self.color, str|tuple|list), "'facecolor' must be str|tuple|list"

		self.alpha = params.get("alpha")
		assert isinstance(self.alpha, None|numbers.Real)

		self.style = params.get("hatch")
		if self.style != None:
			assert isinstance(self.style, str), "'hatch' must be str"
		else:
			self.style = "solid"

		_hatches = ["solid", "none", "\\", "x", "/", "+", "-", "|"]
		_codes = [100, 106, 111, 112, 113, 114, 115, 116]

		index = _hatches.index(self.style) if self.style in _hatches else -1
		if index<0:
			raise ValueError("hatch types: " + ",".join(_hatches))
		else:
			self.style = _codes[index] #integer value for wxWidgets
	

	def __iter__(self):
		return iter([
			("style",self.style),
			("color", self.color),
			("alpha", self.alpha)])





class Font:
	def __init__(self, *args):
		assert isinstance(args[0], dict), "first argument must be dict"
		params:dict = args[0]
		
		self.size = params.get("fontsize")
		if self.size != None:
			assert isinstance(self.size, int), "fontsize must be int"
			assert self.size>0, "size>0 expected"
		else:
			self.size = 11
		
		self.facename = params.get("fontname")
		if self.facename != None:
			assert isinstance(self.facename, str), "fontname must be str"
		else:
			self.facename = "Arial"

		self.weight:str = params.get("fontweight")
		if self.weight == None:
			self.weight = "normal"
		else:
			_weights = ["normal", "light", "bold", "heavy", "ultrabold"]
			if not self.weight in _weights:
				raise ValueError("weights: " + ",".join(_weights))
		
		self.style = params.get("fontstyle")
		if self.style == None:
			self.style = "normal"
		else:
			_styles = ["normal", "italic", "oblique"]
			if not self.style in _styles:
				raise ValueError("weights: " + ",".join(_styles))
		
	
	def __iter__(self):
		return iter([
			("style",self.style),
			("facename", self.facename),
			("weight", self.weight),
			("size", self.size)
		])


import ctypes as _ct
import math
import numbers
from typing import Iterable as _Iterable

from .._ctypeslib import pydll as _pydll
from .gdiobj import Pen, Brush, Font




def text(
		xy:tuple|list, 
		label:str,
		rotation:float = 0.0,
		**kwargs)->None:
	"""
	xy: 	(x, y), top-left,
	label: 	text to be drawn,
	rotation: rotation angle (>0 is counterclockwise; the full angle is 360 degrees)
	labelcolor: label color RGB as str|tuple|list
	"""
	assert isinstance(xy, tuple|list), "xy must be tuple|list"
	assert isinstance(label, str), "label must be string"
	assert isinstance(rotation, numbers.Real), "rotation must be real number"

	_p1 = [i for i in xy if isinstance(i, numbers.Real)]
	assert len(_p1) == 2, "p must contain exactly two real numbers"
	
	_color = kwargs.get("labelcolor") or "0 0 0"
	assert isinstance(_color, str|tuple|list), "color must be str|tuple|list"

	if isinstance(_color, tuple|list):
		_color = ",".join(_color)

	_pydll.c_plot_gdi_text(
			_ct.c_double(xy[0]),
			_ct.c_double(xy[1]),
			_ct.c_char_p(label.encode()),
			_ct.c_double(rotation),
			_ct.c_char_p(_color.encode()),
			dict(Font(kwargs)))



def marker(
		xy:tuple|list, 
		type:str = "c",
		size:int = 5,
		**kwargs)->None:
	"""
	`xy:`	(x, y), centroid,
	`type:`	type of the marker, "c", "t", "r", "x",
	`size:`	size of the marker in pixels
	"""

	assert isinstance(xy, tuple|list), "xy must be tuple|list"
	assert isinstance(type, str), "type must be string"
	assert isinstance(size, int), "size must be int"

	assert 1<size<=100, "1 < size <= 100 expected"

	_p1 = [i for i in xy if isinstance(i, numbers.Real)]
	assert len(_p1) == 2, "p must contain exactly two real numbers"

	_pydll.c_plot_gdi_marker(
			_ct.c_double(xy[0]),
			_ct.c_double(xy[1]),
			_ct.c_char_p(type.encode()),
			_ct.c_uint8(size),
			dict(Pen(kwargs)),
			dict(Brush(kwargs)))



def arc(
		center:tuple|list, 
		p1:tuple|list, 
		p2:tuple|list,
		**kwargs)->None:
	"""
	`center:` (x, y) -> center point of arc
	`p1:` (x1, y1) -> start of arc
	`p2:` (x2, y2) -> end of arc

	## Note:
	To be able to plot a circular arc, the plot area must be a square.
	"""


	assert isinstance(center, tuple|list), "center must be tuple|list"
	assert isinstance(p1, tuple|list), "p1 must be tuple|list"
	assert isinstance(p2, tuple|list), "p2 must be tuple|list"

	_c = [i for i in center if isinstance(i, numbers.Real)]
	assert len(_c) == 2, "center must contain exactly two real numbers"

	_p1 = [i for i in p1 if isinstance(i, numbers.Real)]
	assert len(_p1) == 2, "p1 must contain exactly two real numbers"

	_p2 = [i for i in p2 if isinstance(i, numbers.Real)]
	assert len(_p2) == 2, "p2 must contain exactly two real numbers"

	_pydll.c_plot_gdi_arc(
			_ct.c_double(p1[0]),
			_ct.c_double(p1[1]),
			_ct.c_double(p2[0]),
			_ct.c_double(p2[1]),
			_ct.c_double(center[0]),
			_ct.c_double(center[1]),
			dict(Pen(kwargs)),
			dict(Brush(kwargs)))



def arrow(
		p1:tuple|list, 
		p2:tuple|list, 
		angle:numbers.Real = 45, #45 degrees
		length:float = 0.1, #10% length of main-line
		**kwargs)->None:
	"""
	`p1, p2:` (x1, y1), (x2, y2) coordinate of the main-line
	`angle:` angle between the two head-lines
	`length:` ratio of the length of the head-line to the main-line
	"""

	assert isinstance(p1, tuple|list), "p1 must be tuple|list"
	assert isinstance(p2, tuple|list), "p2 must be tuple|list"
	assert isinstance(angle, numbers.Real), "angle must be Real"
	assert isinstance(length, float), "length must be float"

	_p1 = [i for i in p1 if isinstance(i, numbers.Real)]
	assert len(_p1) == 2, "p1 must contain exactly two real numbers"

	_p2 = [i for i in p2 if isinstance(i, numbers.Real)]
	assert len(_p2) == 2, "p2 must contain exactly two real numbers"

	assert 5 < angle <=180, "5 < angle <=180 expected" 
	assert 0.01 < length < 1, "0.01 < length < 1 expected"

	_pydll.c_plot_gdi_arrow(
			_ct.c_double(p1[0]),
			_ct.c_double(p1[1]),
			_ct.c_double(p2[0]),
			_ct.c_double(p2[1]),
			_ct.c_double(angle),
			_ct.c_double(length),
			dict(Pen(kwargs)))



def curve(
		x: _Iterable, 
		y:_Iterable, 
		**kwargs)->None:
	"""
	Draws a smooth curve between (x1, y1), (x2, y2), ..., (xn, yn). 
	The curve is only guaranteed to pass from (x1, y1) and (xn, yn).

	`x:` x values
	`y:` y values
	"""

	assert isinstance(x, _Iterable), "x must be Iterable"
	assert isinstance(y, _Iterable), "y must be Iterable"

	#pre-check
	assert len(x) == len(y), "x and y must have same lengths"

	_x = [i for i in x if isinstance(i, numbers.Real)]
	assert len(_x) >= 3, "x must contain at least 3 real numbers"

	_y = [i for i in y if isinstance(i, numbers.Real)]
	assert len(_y) >= 3, "y must contain at least 3 real numbers"

	#processed-check
	assert len(_x) == len(_y), "x and y must have same lengths"

	_pydll.c_plot_gdi_curve(x, y, dict(Pen(kwargs)))



def ellipse(
		xy:tuple|list, 
		width:numbers.Real, 
		height:numbers.Real, 
		**kwargs)->None:
	"""
	xy:	 	(x, y), center,
	width: 	half width (>0),
	height: half height (>0),
	"""

	assert isinstance(xy, tuple|list), "p must be tuple|list"
	assert isinstance(width, numbers.Real), "width must be real number"
	assert isinstance(height, numbers.Real), "height must be real number"
	
	assert width>0, "width>0 expected"
	assert height>0, "height>0 expected"

	_p1 = [i for i in xy if isinstance(i, numbers.Real)]
	assert len(_p1) == 2, "p must contain exactly two real numbers"

	if kwargs.get("hatch") == None:
		kwargs["hatch"] = "none"

	_pydll.c_plot_gdi_ellipse(
			_ct.c_double(xy[0]),
			_ct.c_double(xy[1]),
			_ct.c_double(width),
			_ct.c_double(height),
			dict(Pen(kwargs)),
			dict(Brush(kwargs)))	



def line(
		p1:tuple|list, 
		p2:tuple|list, 
		label:str = "",
		**kwargs)->None:
	"""
	`p1:` (x1, y1)
	`p2:` (x2, y2)
	`label:` a text to be shown along with line

	## Note:
	If label is specified:
	`labeldist:float:` (0, 1) pos = start + labeldist*length
	`labelcolor:str|tuple|float` label color
	"""

	assert isinstance(p1, tuple|list), "p1 must be tuple|list"
	assert isinstance(p2, tuple|list), "p2 must be tuple|list"


	_p1 = [i for i in p1 if isinstance(i, numbers.Real)]
	assert len(_p1) == 2, "p1 must contain exactly two real numbers"

	_p2 = [i for i in p2 if isinstance(i, numbers.Real)]
	assert len(_p2) == 2, "p2 must contain exactly two real numbers"


	_pydll.c_plot_gdi_line(
			_ct.c_double(p1[0]),
			_ct.c_double(p1[1]),
			_ct.c_double(p2[0]),
			_ct.c_double(p2[1]),
			dict(Pen(kwargs)))
	
	_txt = label.rstrip()
	_txt = _txt.lstrip()

	if _txt != "":
		_labelcolor = kwargs.get("labelcolor") or "0 0 0"
		_labeldist = kwargs.get("labeldist") or 0.45
		assert isinstance(_labeldist, float), "labeldist must be float"
		assert isinstance(_labelcolor, str|tuple|list), "labelcolor must be str|tuple|list"

		dy = p2[1] - p1[1]
		dx = p2[0] - p1[0]
		Slope = math.atan2(dy, dx)

		TL = (
				p1[0] + _labeldist * math.cos(Slope) ,
				p1[1] + _labeldist * math.sin(Slope))

		
		#As of this point slope is in degrees (0, 180) or (0, -180)
		Slope = Slope*180/math.pi

		if Slope<0:
			Slope += 180

		if Slope>90:
			Slope = 90 - Slope
	
		text(TL, label, Slope, **kwargs)



def polygon(
		x: _Iterable, 
		y:_Iterable, 
		**kwargs)->None:
	"""
	Draws a polygon between (x1, y1), (x2, y2), ..., (xn, yn). 
	The first and last points are automatically closed.

	`x:` x values
	`y:` y values
	"""

	assert isinstance(x, _Iterable), "x must be Iterable"
	assert isinstance(y, _Iterable), "y must be Iterable"

	#pre-check
	assert len(x) == len(y), "x and y must have same lengths"

	_x = [i for i in x if isinstance(i, numbers.Real)]
	assert len(_x) >= 3, "x must contain at least 3 real numbers"

	_y = [i for i in y if isinstance(i, numbers.Real)]
	assert len(_y) >= 3, "y must contain at least 3 real numbers"

	#processed-check
	assert len(_x) == len(_y), "x and y must have same lengths"

	_pydll.c_plot_gdi_polygon(
			x, y, 
			dict(Pen(kwargs)), 
			dict(Brush(kwargs)))




def rect(
		xy:tuple|list, 
		width:numbers.Real, 
		height:numbers.Real, 
		**kwargs)->None:
	"""
	xy:	(x, y), bottom-left corner of the rectangle,
	width: 	width of rectangle (>0),
	height: height of rectangle (>0), 
	"""	

	assert isinstance(xy, tuple|list), "xy must be tuple|list"
	assert isinstance(width, numbers.Real), "width must be real number"
	assert isinstance(height, numbers.Real), "height must be real number"

	assert width>0, "width>0 expected"
	assert height>0, "height>0 expected"

	_p1 = [i for i in xy if isinstance(i, numbers.Real)]
	assert len(_p1) == 2, "p must contain exactly two real numbers"

	kwargs["hatch"] = kwargs.get("hatch") or "none" #make it transparent by default

	_pydll.c_plot_gdi_rect(
			_ct.c_double(xy[0]),
			_ct.c_double(xy[1]),
			_ct.c_double(width),
			_ct.c_double(height),
			dict(Pen(kwargs)),
			dict(Brush(kwargs)))



from ._ctypeslib import pydll as _pydll
import ctypes as _ct
import dataclasses as _dc
import numbers as _numbers
from numpy.polynomial import polynomial as _Polynomial
import numpy as _np
from typing import Iterable



__all__ = [
	'approx',
	'linearinterp', 
	'lagrange', 
	'spline', 
	'expfit', 
	'logfit', 
	'logistfit', 
	'polyfit', 
	'powfit', 
	'SplineResult']

def _MinMax(X:_np.ndarray)->tuple[float]:
	_min = X[0]
	_max = X[0]
	for i in range(1, len(X)):
		_min = X[i] if X[i]<_min else _min
		_max = X[i] if X[i]>_max else _max
	
	return _min, _max




def linearinterp(x1:float, y1:float, x2:float, y2:float, xval:float)->float:
	"""
	Linear interpolation

	## Inputs:
	x1, y1: First point \n
	x2, y2: Second point \n
	xval: x-value in [x1, x2] 

	## Returns:
	y-value in [y1, y2] corresponding to xval
	"""

	if(x1 == x2): 
		return y1 
	
	m,n = 0, 0
	m = (y2 - y1) / (x2 - x1)
	n = y2 - m * x2

	return m * xval + n





def lagrange(
		x:Iterable, 
		y:Iterable, 
		value:float)->float:
	"""
	Constructs lagrange polynomial from x,y to compute the given value.
	"""
	assert issubclass(value, _numbers.Real)

	return _pydll.c_fit_lagrange(x, y, _ct.c_double(value))


########################################################################################


@_dc.dataclass
class SplineResult:
	"""
	poly: Numpy polynomial
	lower, upper: lower and upper bounds of the polynomial
	"""
	poly:_Polynomial.Polynomial = None
	lower:float = None
	upper: float = None


def spline(x:Iterable, y:Iterable)->list[SplineResult]:
	"""
	Constructs natural cubic spline polynomials from x, y
	"""
	lst = _pydll.c_fit_spline(x, y)
	
	return  [SplineResult(_Polynomial.Polynomial(l[0]), l[1], l[2]) for l in lst]


########################################################################################


def expfit(
		x:Iterable, 
		y:Iterable, 
		intercept:None|float=None)->list:
	"""
	Fits x,y to the equation y = a*exp(b*x) \n
	Returns [a, b]

	## Note:
	If intercept is not provided both a, b are computed. \n
	If intercept is given then a=intercept and b is computed accordingly.
	"""
	if(intercept!=None):
		assert issubclass(intercept, _numbers.Real)

	return _pydll.c_fit_expfit(x, y, intercept)


def logfit(x:Iterable, y:Iterable)->list:
	"""
	Fits x,y to the equation y = a*ln(x) + b \n
	Returns [a, b]
	"""

	return _pydll.c_fit_logfit(x, y)


def logistfit(x:Iterable, y:Iterable, limit = None)->list:
	"""
	Fits to the equation y = L / (1 + exp(b0 + b1*x)) \n
	Returns either [L, b0, b1] or [b0, b1]

	## Inputs:
	limit: None or float
	"""
	if(limit!=None):
		assert issubclass(limit, _numbers.Real)

	return _pydll.c_fit_logistfit(x, y, limit)


def polyfit(x:Iterable, y:Iterable, deg)->tuple:
	"""
	Uses numpy.polynomial.polynomial.polyfit
	returns (coefficients, residuals)
	"""
	coeffs, stats = _Polynomial.polyfit(x=x, y=y, deg=deg)
	return coeffs, stats[0]


def powfit(x:Iterable, y:Iterable)->list[float]:
	"""
	Fits to the equation y = a*x^n \n
	Returns [a, n]
	"""
	return _pydll.c_fit_powfit(x,y)




def approx(
		x:_np.ndarray, 
		y:_np.ndarray, 
		n=50)->tuple[_np.ndarray, _np.ndarray]:
	"""
	Returns points which linearly interpolate (
	interpolation takes place at n equally spaced points 
	spanning the interval [min(x), max(x)]).

		
	`x, y:` Numeric vectors giving the coordinates of the points to be interpolated. 
	`n:` Number of equally spaced data points [min(x), min(y)]
	"""
	assert n>1, "n>1 expected"

	v = _np.zeros(n)
	if type(y) == type(None):
		Min, Max = _MinMax(x)
		strideLen = (Max-Min)/(n-1)
		v[0], v[n-1] = Min, Max
		for i in range(1, len(v)-1):
			v[i] = Min + i * strideLen
		
		return v
	
	assert len(x)==len(y), "x and y must have same lengths"

	"""
	Generate n data points in the interval min(x) and max(x)
	Note that data points are sorted by nature.
	"""
	XX = approx(x, None, n)

	for i in range(len(x)-1):
		for j in range(n):
			if (x[i] <= XX[j] <= x[i + 1]):
				x1, x2 = x[i], x[i + 1]
				y1, y2 = y[i], y[i + 1]
				v[j] = linearinterp(x1, y1, x2, y2, XX[j])

	return XX, v


import math as _math

class Food:
	pass



def ComputeAw_T(food:Food, aw1:float)->float|None:
	"""
	Computes aw at a different temperatures
	
	food: Food material
	aw1: water activity of food at reference temperature (generally around 20C)
	T2: the temperature at which water activity will be computed
	"""
	Tref = 20
	if _math.isclose(food.T, Tref, abs_tol=1.0):
		return aw1

	#save the actual temperature
	T = food.T

	food.T = Tref
	Cp_20 = food.cp()

	#restore the actual temperature
	food.T = T
	Cp_T = food.cp()

	Cp_avg = (Cp_20 + Cp_T) / 2.0	
	Qs = food.molecularweight()* Cp_avg*(T - 20.0) #kJ/kg

	R = 8.314 #kPa*m^3/kgK

	T += 273.15
	dT = 1/293.15 - 1/T

	aw2 = aw1*_math.exp(Qs/R*dT)
	
	return aw2 if aw2>=0 and aw2<=1 else None



#-----------------------------------------------------------------------------


class Aw():
	def __init__(self, food:Food) -> None:
		self._food = food


	def FerroFontan_Chirife_Boquet(self)->float:
		"""
		Assumptions: 
		CHO (fructose), protein (alanine), lipid (glycerol)
		"""
		MW_CHO = 180.16 #fructose
		MW_LIPID = 92.0944 #glycerol
		MW_PROTEIN = 89.09 #alanine

		fd = self._food

		NCHO, NLipid, NProtein = fd.cho/MW_CHO, fd.lipid/ MW_LIPID, fd.protein/MW_PROTEIN
		
		#number of moles of water
		N_w = fd.water/18.02

		#solute
		N_slt = NCHO + NLipid + NProtein

		Solute = 1 - fd.water
		C_cho, C_Lipid, C_Protein = fd.cho/Solute, fd.lipid/Solute, fd.protein/Solute

		Mt = _math.sqrt(C_cho/MW_CHO + C_Lipid/MW_LIPID + C_Protein/MW_PROTEIN)

		# Norrish equation K values using Ferro-Chirife-Boquet equation
		Km = C_cho*(Mt/MW_CHO)*(-2.15) +C_Lipid*(Mt/MW_LIPID)*(-1.16) + C_Protein*(Mt/MW_PROTEIN)*(-2.52) 
		
		# Mole fraction of solute
		X_slt = N_slt/(N_slt + N_w) 

		# Mole fraction of water
		XWater = N_w/(N_slt + N_w) 

		aw = XWater*_math.exp(Km*X_slt**2)

		return aw



	def Norrish(self)->float:
		"""Norrish equation"""
		
		f = self._food

		#CHO is considered as fructose
		N_cho = f.cho/180.16 
		
		#lipid is considered as glycerol
		N_l = f.lipid/ 92.0944 
		
		#protein is considered as alanine
		N_p = f.protein/89.09 

		#water
		N_w = f.water / 18

		#total
		N_tot = N_cho + N_l + N_p + N_w

		#X_l: lipid, X_p: protein, X_w: water
		X_cho, X_l, X_p, X_w = N_cho/N_tot, N_l/N_tot, N_p/N_tot, N_w/N_tot

		
		# Norrish equation K values using Ferro-Chirife-Boquet equation
		SumSq = -(2.15)*X_cho**2 - (1.16)*X_l**2 - (2.52)*X_p**2
		SumX2 = X_cho**2 + X_l**2 + X_p**2
		
		rhs = _math.log(X_w) + SumSq/SumX2*(1-X_w)**2

		return _math.exp(rhs)



	#mostly used in confectionaries
	def MoneyBorn(self)->float:
		f = self._food

		# amount of CHO in 100 g water (equation considers thus way) 
		W = 100*f.cho/f.water 
		
		#CHO is considered as fructose
		N_cho = W/180.16 
		
		return 1.0/(1.0 + 0.27*N_cho)



	def Raoult(self)->float:
		#prediction using Raoult's law
		#all in percentages

		f = self._food

		x_w = f.water
		xCHO =  f.cho
		x_l = f.lipid
		x_p =  f.protein

		#solute
		x_slt = xCHO + x_l + x_p + f.ash

		#average molecular weight
		MW_slt = (xCHO/x_slt)*180.16 + (x_l/x_slt)*92.0944 + (x_p/x_slt)*89.09 if x_slt>0 else 1
			
		MW_w=18 #molecular weight of water
		MW_nacl = 58.44

		temp1 = x_w + (MW_w/MW_slt)*x_slt + 2*(MW_w/MW_nacl)*f.salt
		return x_w/temp1





""" ------------------------------------------------------------------------------- """

class Cp():
	def __init__(self, food:Food) -> None:
		self._food = food

	def Siebel(self, Tf = -1.7)->float:
		"""
		returns kJ/kg°C \n

		## Input:
		Tf = -1.7 is the default freezing temperature

		## Reference:
		Siebel, E (1892). Specific heats of various products. Ice and Refrigeration, 2, 256-257.
		"""

		food = self._food

		Fat = food.lipid
		SNF = food.ash + food.protein + food.cho
		M = food.water
		Tfood = food.T


		#for fat free foods
		if _math.isclose(Fat, 0.0, abs_tol=1E-5):
			r = 837.36
			r += 3349*M if Tfood>Tf else 1256*M
			return r/1000

		r = 1674.72*Fat +  837.36*SNF
		r += 4186.8*M if Tfood>Tf else 2093.4*M

		return r/1000


	def Heldman(self)->float:
		"""
		returns kJ/kg°C 

		## Reference:
		Heldman, DR (1975). Food Process Engineering. Westport, CT: AVI 
		"""
		food = self._food

		Fat = food.lipid
		Protein = food.protein
		Ash = food.ash
		cho = food.cho
		water = food.water
		
		return 4.18*water + 1.547*Protein + 1.672*Fat + 1.42*cho + 0.836*Ash


	def Chen(self)->float:
		"""
		specific heat of an unfrozen food returns kJ/kg°C \n

		## Reference:
		Chen CS (1985). Thermodynamic Analysis of the Freezing and Thawing of Foods: 
		Enthalpy and Apparent Specific Heat. Food Science, 50(4), 1158-1162
		"""
		food = self._food

		Solid = 1 - food.water	
		return 4.19 - 2.30*Solid - 0.628*Solid**3



import ctypes as _ct

from dataclasses import dataclass as _dataclass





from .core import kurt, mode, moveavg, rolling, sample, skew

from .aov import aov, aov2, aov2_results
from .linregress import linregress

from .test_basic import \
	test_norm_ad, TestNormRes, \
	test_f, test_f_Result, \
	test_sign, test_sign_Result,\
	test_t, test_t1_result, test_t2_result, test_tpaired_result, \
	test_z, test_z_Result


from .distributions import \
	dbeta, pbeta, qbeta, rbeta, \
	dbinom, pbinom, qbinom, rbinom, \
	dchisq, pchisq, qchisq, rchisq, \
	dexp, pexp, qexp, rexp, \
	df, pf, qf, rf, \
	dgamma, pgamma, qgamma,rgamma, \
	dgeom, pgeom, qgeom, rgeom, \
	dhyper, phyper, qhyper, rhyper, \
	dmultinom, rmultinom, \
	dnbinom, pnbinom, qnbinom,rnbinom, \
	dnorm, pnorm, qnorm, rnorm, \
	dlnorm, plnorm, qlnorm, rlnorm, \
	dpois, ppois, qpois, rpois, \
	dsignrank, qsignrank, psignrank, \
	dt, pt, qt, rt, \
	dunif, punif, qunif, runif, \
	dweibull, pweibull, qweibull, rweibull







import math
import numbers

import numpy as np
from .distributions import pf, pt, qt




__all__=['linregress', 'linregressResult']





def FitZeroIntercept(yobs:np.ndarray, factor:np.ndarray):
	"""
	Equation to be solved: a1.x=y1, a2.x=y2 ..... an.x=yn

	Best solution: trans(A)*A*x=trans(A)*b	where A is a matrix with first column a's and second column 0s
	trans(A)*A = a1^2+a2^2+...+an^2
	trans(A)*b = a1*b1+a2*b2+...+an*bn
	x = (trans(A)*b) / (trans(A)*A)

		Also note that (important for coefficient analysis)
	var(beta1) = var(sum(xy)/sum(x^2)) = [1/sum(x^2)]^2 * sum(x^2)*var(population)
	sd(beta1) = S(population)/sqrt(sum(x^2))

	"""
	sum_x2 = 0
	sum_xy = 0

	assert len(yobs) == len(factor), "yobs and factor have different lengths"
	
	for i in range(len(yobs)):
		sum_x2 += factor[i]**2.0
		sum_xy += factor[i]*yobs[i]
		
	return sum_xy/sum_x2




class linregressResult:
	"""
	Do NOT create an instance of this class directly \n
	An instance is returned by simple_linregress or multiple_linregress classes' summary methods
	"""
	def __init__(self, Dict) -> None:
		self.m_Dict = Dict
      
	@property
	def all(self):
		"""returns the dictionary containing all results"""
		return self.m_Dict


	@property
	def R2(self):
		return self.m_Dict["R2"]
	
	@property
	def stderr(self):
		"""standard error"""
		return self.m_Dict["SE"]

	@property
	def pvalue(self):
		"""p-value from ANOVA stat"""
		return self.m_Dict["ANOVA"]["pvalue"]
      

	@property
	def fvalue(self):
		"""F-value from ANOVA stats"""
		return self.m_Dict["ANOVA"]["Fvalue"]
      
     

	@property
	def intercept(self):
		"""
		returns {coeff, pvalue, tvalue, SE, CILow, CIHigh}
		"""
		if(len(self.m_Dict["CoefStats"])>1):
			return self.m_Dict["CoefStats"][0]
		
		return None

	@property
	def ANOVA(self):
		"""
		returns {DF_Residual, SS_Residual, MS_Residual, DF_Regression, SS_Regression, MS_Regression 
		SS_Total, Fvalue, pvalue}
		"""
		return self.m_Dict["ANOVA"]

	@property
	def coeffstat(self):
		"""
		{coeff, pvalue, tvalue, SE, CILow, CIHigh}
		"""
		return self.m_Dict["CoefStats"]





class simple_linregress:
	""" simple linear model """
      
	def __init__(self, yobs, factor, intercept=True, alpha=0.05) -> None:
		assert isinstance(yobs, np.ndarray), "yobs must be of type numpy array"
		assert isinstance(factor, np.ndarray), "factor must be of numpy array"

		self.m_yobs = yobs
		self.m_factor = factor
		self.m_intercept = intercept
		self.m_alpha = alpha
		

	def compute(self)->list:
		"""
		returns [slope, [intercept]] and must be called before summary()
		"""
		if(self.m_intercept):
			self.m_coeffs = np.polyfit(self.m_factor, self.m_yobs, 1)
		else:
			self.m_coeffs = np.zeros(2)
			self.m_coeffs[0] = FitZeroIntercept(self.m_yobs, self.m_factor)
		
		return self.m_coeffs.tolist()




	def __str__(self) -> str:
		assert len(self.m_coeffs) > 0, "compute must be called first"

		retStr=""
		if(self.m_intercept):
			retStr += str(self.m_coeffs[0])+"*x + " + str(self.m_coeffs[1])
		else:
			retStr += str(self.m_coeffs[0])+"*x" 
		
		return retStr




	def summary(self):
		assert self.m_coeffs.size > 0, "compute must be called first"
		
		N = self.m_yobs.size
		assert N >= 3, "At least 3 entries must be provided"
		
		mean_x, mean_y = np.mean(self.m_factor), np.mean(self.m_yobs)
		sum_xy, sum_x2, sum_y2, sum_y, sum_x, sum_mean_x, SS_Total = 0, 0, 0, 0, 0, 0, 0

		for i in range(N):
			xi, yi = float(self.m_factor[i]), float(self.m_yobs[i])
			sum_x += xi
			sum_y += yi
			sum_x2 += xi**2
			sum_y2 += yi**2
			sum_xy += xi*yi
			sum_mean_x += (xi - mean_x)**2
			SS_Total += (yi - mean_y)**2 #total variability (SS total)


		df = N-2
		if(self.m_intercept == False):
			df = N-1
			SS_Total=sum_y2
            
		#Forming the ANOVA table for regression
		fit_y = np.polyval(self.m_coeffs, self.m_factor)
		residual = self.m_yobs - fit_y

		SS_Residual = float(sum(residual**2))
		SS_Regression = SS_Total - SS_Residual

		MS_Regression, MS_Residual = SS_Regression, SS_Residual/df

		ANOVA = {"DF_Residual":df, "SS_Residual":SS_Residual, "MS_Residual":MS_Residual,
			"DF_Regression":1, "SS_Regression":SS_Regression, "MS_Regression":MS_Regression, 
			"SS_Total":SS_Total}
		
		ANOVA["Fvalue"] = float(MS_Regression/MS_Residual)
		ANOVA["pvalue"] = 1 - pf(float(MS_Regression/MS_Residual), 1, df)



		def CoeffStat(beta, SE_beta, t_beta):
			pvalue=0

			#area on the left of tcrit + area on the right of positive
			if(t_beta<=0):
				pvalue = pt(q=t_beta, df=df) + (1-pt(q=abs(t_beta), df=df))
			
			#area on the right of positive tcritical + area on the left of negative tcritical
			elif(t_beta>0):
				pvalue = (1-pt(q=t_beta, df=df)) + pt(q=-t_beta, df=df) 
			

			tbl = {"coeff":beta, "pvalue":pvalue, "tvalue":t_beta, "SE":SE_beta }

			invTval = qt(self.m_alpha/2.0, ANOVA["DF_Residual"])

			val1 = beta - SE_beta*invTval
			val2 = beta + SE_beta*invTval

			tbl["CILow"] = min(val1,val2)
			tbl["CIHigh"] = max(val1,val2)

			return tbl


		#std error of population
		s = math.sqrt(MS_Residual)

		SE_beta1=s/math.sqrt(sum_mean_x)

		if(not self.m_intercept):
			SE_beta1 = s/math.sqrt(sum_x2)
		
		beta1, beta0 = float(self.m_coeffs[0]), float(self.m_coeffs[1])
		t_beta1 = beta1/SE_beta1

		tbl1 = CoeffStat(beta1, SE_beta1, t_beta1)

		CoefStats=[]


		#tbl0
		if(self.m_intercept):
			SE_beta0 = s*math.sqrt(sum_x2)/(math.sqrt(N)*math.sqrt(sum_mean_x))
			t_beta0 = beta0/SE_beta0
			
			tbl0 = CoeffStat(beta0, SE_beta0, t_beta0)

			CoefStats=[tbl0, tbl1]
		else:
			CoefStats=[tbl1]


		R2=SS_Regression/SS_Total

		retTable={"CoefStats":CoefStats, "ANOVA":ANOVA, "R2":R2, "SE":s}

		ResultClass = linregressResult(retTable)

		return ResultClass


	def residuals(self)->tuple:
		IsIntercept = self.m_intercept
		Coeffs = self.m_coeffs

		Residuals = []
		Fits = []

		NRows = self.m_factor.shape

		for i in range(NRows):
			Intercept = float(Coeffs[0]) if IsIntercept else 0.0
			fit = Intercept
			
			Coeff = float(Coeffs[1]) if IsIntercept else float(Coeffs[0])
			fit += self.m_factor[i]*Coeff	
			
			residual = self.m_yobs[i] - fit
			
			Residuals.append(float(residual))
			Fits.append(float(fit))
		
		return (Residuals, Fits)





class multiple_linregress:
	"""
	multiple linear regression
	"""
	def __init__(self, yobs:np.ndarray, factor:np.ndarray, intercept=True, alpha=0.05) -> None:
		assert isinstance(yobs, np.ndarray), "yobs must be of type numpy 1D array"
		assert isinstance(factor, np.ndarray), "factor must be of type numpy 2D array"

		NRows = factor.shape[0]
		assert NRows == yobs.size, "N rows of matrix must be equal to the size of the observed variables."

		self.m_yobs = yobs
		self.m_factor = factor
		self.m_intercept = intercept
		self.m_alpha = alpha



	def compute(self)->list:
		self.m_modifiedMatrix = np.copy(self.m_factor)

		if(self.m_intercept):
			NRows = self.m_factor.shape[0]
			ones = np.ones(NRows)
			self.m_modifiedMatrix = np.insert(self.m_modifiedMatrix, 0, values=ones, axis=1)

		self.m_coeffs = np.linalg.lstsq(a=self.m_modifiedMatrix, b=self.m_yobs, rcond=None)[0]

		return self.m_coeffs.tolist()


	def __str__(self) -> str:
		""" returns as a0 + a1*X1 + a2*X2 + ... """
		assert len(self.m_coeffs) > 0, "compute must be called first"

		retStr=""
		N = len(self.m_coeffs)
		
		if(self.m_intercept):
			retStr += str(self.m_coeffs[0]) + " + "
			
			for i in range(1, N-1):
				retStr += str(self.m_coeffs[i]) + "*x" + str(i) + " + "
			
			retStr += str(self.m_coeffs[N-1]) + "*x" + str(N-1)
		else:
			for i in range(0, N-1):
				retStr += str(self.m_coeffs[i]) + "*x" +str(i+1) + " + "
			
			retStr += str(self.m_coeffs[N-1]) + "*x" +str(N)
		
		return retStr


	def summary(self):
		assert len(self.m_coeffs) > 0, "compute must be called first"

		nrows, ncols = self.m_factor.shape

		MeanYObs = np.mean(self.m_yobs)
		ypredicted = np.zeros(nrows)

		SS_Total, SS_Residual = 0, 0
		sum_y2=0
		for i in range(nrows):
			#row vector * col vector = number
			ypredicted[i] = np.dot(self.m_modifiedMatrix[i, :], self.m_coeffs)
			
			SS_Total += (MeanYObs-self.m_yobs[i])**2
			SS_Residual += (self.m_yobs[i]-ypredicted[i])**2
			sum_y2 += self.m_yobs[i]**2
		
		NObservations = nrows
		DF_Regression = ncols
		DF_Residual = NObservations - (DF_Regression + 1)
		if(not self.m_intercept):
			DF_Residual = NObservations - DF_Regression
			SS_Total = sum_y2
		
		DF_Total = DF_Regression + DF_Residual

		SS_Regression = SS_Total - SS_Residual

		R2 = SS_Regression/SS_Total

		MS_Residual = SS_Residual / DF_Residual
		MS_Regression = SS_Regression/DF_Regression

		FValue = MS_Regression/MS_Residual

		pvalue = 1 - pf(float(FValue), DF_Regression, DF_Residual)

		ANOVA = {"SS_Total":SS_Total, "SS_Residual":SS_Residual, "SS_Regression":SS_Regression,
				"DF_Regression":DF_Regression, "DF_Residual":DF_Residual, "MS_Residual":MS_Residual,
				"MS_Regression": MS_Regression, 
				"Fvalue": FValue,
				"pvalue": pvalue}
		

		SEMat = np.linalg.inv(np.dot(np.transpose(self.m_modifiedMatrix),self.m_modifiedMatrix))
		SE = np.sqrt(np.diag(SEMat))*math.sqrt(MS_Residual)

		CoefStats=[]
		for i in range(len(self.m_coeffs)):
			tbl = {}

			tbl["coeff"] = self.m_coeffs[i]
			tbl["SE"] = SE[i]
			
			Tvalue = self.m_coeffs[i]/SE[i]
			tbl["tvalue"] = Tvalue

			if(Tvalue>=0):
				tbl["pvalue"] = 2*(1 - pt(q=float(Tvalue), df = nrows-1))
			else:
				tbl["pvalue"] = 2*pt(q=float(Tvalue), df = nrows-1)
			
			
			invTval = qt(self.m_alpha/2.0, DF_Residual)
			
			val1 = self.m_coeffs[i] - SE[i]*invTval
			val2 = self.m_coeffs[i] + SE[i]*invTval

			tbl["CILow"] = min(val1,val2)
			tbl["CIHigh"] = max(val1,val2)

			CoefStats.append(tbl)


		retTable = {"CoefStats":CoefStats, "ANOVA":ANOVA, "R2":R2, "SE":SE}

		return linregressResult(retTable)


	def residuals(self)->tuple:
		IsIntercept = self.m_intercept
		Coeffs = self.m_coeffs

		Residuals = []
		Fits = []

		NRows, NCols = self.m_factor.shape

		for i in range(NRows):
			Intercept = float(Coeffs[0]) if IsIntercept else 0.0
			fit = Intercept
			for j in range(NCols):
				Coeff = float(Coeffs[j+1]) if IsIntercept else float(Coeffs[j])
				fit += self.m_factor[i, j]*Coeff	
			
			residual = self.m_yobs[i] - fit
			
			Residuals.append(float(residual))
			Fits.append(float(fit))
		
		return (Residuals, Fits)




def linregress(yobs, factor, intercept=True, alpha=0.05):
	"""
	Performs simple/multiple linear regression

	## Input
	yobs: Dependent data (ndarray|list) \n
	factor: independent data (2D list| 2D ndarray) \n
	intercept: True if there is intercept \n
	alpha: significance level
	"""
	Observed = yobs
	Factor = factor

	if(isinstance(yobs, list)):
		Observed = np.asfarray(yobs)

	if(isinstance(factor, list)):
		Factor = np.asfarray(factor)

	IsMatrix = len(Factor.shape) == 2
	IsVector = len(Factor.shape) == 1

	if(IsMatrix):
		return multiple_linregress(Observed, Factor, intercept, alpha)
	
	elif(IsVector):
		return simple_linregress(Observed, Factor, intercept, alpha)
	
	else:
		return TypeError("factor must be either list, 2D list, numpy.ndarray")


import numpy as np
import numbers
from typing import Iterable as _Iterable

from .gdi import marker
from .charts import canvas
from ..util import NiceNumbers, minmax


def _ComputeAlphas(s:_Iterable)->list[float]:
	"""
	size of the bubble and its alpha are inversely proportional
	e.g., large bubbles have small alphas wherease 
	small bubbles have large alphas
	"""

	Alpha_Min, Alpha_Max = 0.1, 1.75
	Min, Max = minmax(s)

	assert Max>Min, "max > min expected for bubble size"

	Slope = (Alpha_Min - Alpha_Max)/(Max-Min)

	#equation of the line (x: bubble size, y: alpha)
	Alpha = lambda size: Slope*(size-Min) + Alpha_Max

	return [Alpha(i) for i in s]



def _ComputeScaledSizes(s:_Iterable, scale=1.0)->list[float]:
	"""
	Area based computation
	"""
	MINSIZE, MAXSIZE = 5, 50
	xx = np.array(s)
	MaxSize:float = np.max(xx)
	Arr = (xx/MaxSize)**0.5*MAXSIZE*scale
	return Arr.tolist()
	

def bubble(
		x:_Iterable,
		y:_Iterable,  
		s:_Iterable,
		scale:float=1.0,
		color:str|tuple|list = (255, 0, 0),
		label:str=None):
	"""
	Plots bubble chart

	## Input:
	x, y, s:	x- and y- and size data
	color: color
	label: Name (currently not in use)
	"""
	assert \
		isinstance(x, _Iterable) and \
		isinstance(y, _Iterable) and \
		isinstance(s, _Iterable), "x, y and s must be iterable objects."

	assert len(x) == len(y) and len(y) == len(s), "x, y and s must have same lengths"

	assert isinstance(label, str|None), "'label' must be string"
	assert isinstance(color, str|tuple|list), "'color' must be str|tuple|list"

	assert isinstance(scale, numbers.Real), "scale must be Real"
	assert 0<scale<2.0, "scale must be in [0, 2]"


	Alphas = _ComputeAlphas(s)
	Sizes = _ComputeScaledSizes(s, scale)
	MinX, MaxX = minmax(x)
	MinY, MaxY = minmax(y)

	NiceX, NiceY = NiceNumbers(MinX, MaxX), NiceNumbers(MinY, MaxY)

	canvas(
		[NiceX.min - NiceX.tickspace, NiceX.max + NiceX.tickspace], 
		[NiceY.min - NiceY.tickspace, NiceY.max + NiceY.tickspace])
	
	
	"""
	alpha values we use are actually luminiscence values

	If we draw the small bubble before a big bubble even though
	the big bubble has a lower alpha value than the small bubble, 
	the small bubble will not be shown. 
	
	Therefore, we sort in descending order according to bubble sizes and draw 
	the larger ones first.
	"""
	
	Sorted = []
	for i in range(len(x)):
		Sorted.append([x[i], y[i], round(Sizes[i]), Alphas[i]])
	
	#in-place sort (descending order) according to sizes
	Sorted.sort(key=lambda x: x[2], reverse=True)

	for l in Sorted:
		marker(xy=[l[0], l[1]], size=l[2], ec=color, fc=color, alpha=l[3])


import numpy as _np
import numbers

from .charts import canvas
from .gdi import arrow
from ..util import NiceNumbers



def _MinMax(X:_np.ndarray)->tuple[float]:
	_min = X[0]
	_max = X[0]
	for i in range(1, len(X)):
		_min = X[i] if X[i]<_min else _min
		_max = X[i] if X[i]>_max else _max
	
	return _min, _max




def _ComputeScaleFactor(
		X:_np.ndarray, 
		Y:_np.ndarray,
		U:_np.ndarray, 
		V:_np.ndarray)->float:
	minX, maxX = _MinMax(X)
	minY, maxY = _MinMax(Y)

	LenArrows = _np.sqrt(U**2 + V**2)
	MaxLen = _np.max(LenArrows)

	niceX = NiceNumbers(minX, maxX)
	niceY = NiceNumbers(minY, maxY)

	return min(niceX.tickspace, niceY.tickspace)/MaxLen



def quiver(
		X:_np.ndarray, 
		Y:_np.ndarray,
		U:_np.ndarray, 
		V:_np.ndarray,
		scale=None,
		**kwargs)->None:
	
	""""
	Plots quiver chart

	## Input:
	x, y: (x,y) location, 2D ndarray \n
	u, v: length in x and y directions, 2D ndarray \n
	scale: if None automatic scaling takes place otherwise xu = x + u*scale is computed
	"""	
	assert len(X)==len(Y), "X and Y must have same lengths"
	assert len(U) == len(V), "U and V must have same lengths"
	assert len(X) == len(U), "X, Y, U, V must have same lengths"
	assert isinstance(scale, numbers.Real|None), "scale must be float|int|None"

	x, y = X.flatten(), Y.flatten()
	u, v = U.flatten(), V.flatten()

	_scale = scale
	if _scale == None:
		_scale = _ComputeScaleFactor(x, y, u, v)
	
	xu = x + u*_scale
	yv = y + v*_scale

	BndScaledX, BndScaledY = _MinMax(xu), _MinMax(yv)
	BndRawX, BndRawY = _MinMax(x), _MinMax(y)

	_MinX = min(BndScaledX[0], BndRawX[0])
	_MaxX = max(BndScaledX[1], BndRawX[1])
	_MinY = min(BndScaledY[0], BndRawY[0])
	_MaxY = max(BndScaledY[1], BndRawY[1])
	
	niceX = NiceNumbers(_MinX, _MaxX)
	niceY = NiceNumbers(_MinY, _MaxY)
	canvas(niceX.minmax, niceY.minmax)

	for i in range(len(x)):
		arrow(p1=(x[i], y[i]), p2=(xu[i], yv[i]),**kwargs)
	

def dirfield(
		x:_np.ndarray, 
		y:_np.ndarray, 
		slope:_np.ndarray):
	"""
	Plots the direction field for a given function f=dy/dx \n

	## Input
	x, y: 2D numpy array (after using meshgrid) \n
	slope: 2D array resulting from evaluation of f=dy/dx, first order ODE
	"""

	assert isinstance(x, _np.ndarray), "'x' must be ndarray"
	assert isinstance(y, _np.ndarray), "'y' must be ndarray"
	assert isinstance(slope, _np.ndarray), "'slope' must be ndarray"

	# angle of inclination
	t = _np.arctan(slope)

	# xy-components of arrow
	dx = _np.cos(t)
	dy = _np.sin(t); 

	#call quiver to visualize   
	quiver(x, y, dx, dy)




import math
import numbers
from typing import Iterable as _Iterable



def minmax(X:_Iterable)->tuple[numbers.Real]:
	_min = X[0]
	_max = X[0]
	for i in range(1, len(X)):
		_min = X[i] if X[i]<_min else _min
		_max = X[i] if X[i]>_max else _max
	
	return _min, _max




class NiceNumbers:
	def __init__(self, minimum, maximum, maxticks=10) -> None:
		self._min = minimum
		self._max = maximum
		self._maxticks = maxticks

		self._calculate()
	

	def _calculate(self):
		_range = self._nicenum(self._max - self._min, False)
		self._tickspacing = self._nicenum(_range / (self._maxticks - 1), True)

		self._nicemin = math.floor(self._min / self._tickspacing) * self._tickspacing
		self._nicemax = math.ceil(self._max / self._tickspacing) * self._tickspacing


	def _nicenum(self, Range:float, Round:bool):
		exponent = math.floor(math.log10(Range))
		frac = Range / math.pow(10, exponent)

		if Round:
			if frac < 1.5: niceFrac = 1
			elif frac < 3: niceFrac = 2
			elif frac < 7: niceFrac = 5
			else: niceFrac = 10
		else:
			if frac <= 1: niceFrac = 1
			elif frac <= 2: niceFrac = 2
			elif frac <= 5: niceFrac = 5
			else: niceFrac = 10

		return niceFrac * math.pow(10.0, exponent)

	@property
	def min(self):
		return self._nicemin

	@property
	def max(self):
		return self._nicemax
	
	@property
	def minmax(self):
		return self._nicemin, self._nicemax
	
	@property
	def tickspace(self):
		return self._tickspacing
	

"""
if __name__ == "__main__":
	n = NiceNumbers(0.5, 10.25)
	print(n.nicemax, " ", n.nicemin, " ", n.tickspace)
	
"""



from .gdiobj import Pen, Brush


class Marker:
	"""
	A class to define marker properties

	## Input:
	`style:` "c", "s", "t", "x"  \n	
	"""
	def __init__( self, style="c", size=5, **kwargs) -> None:
		_style = style or "c"
		_size = size or 5

		assert isinstance(_style, str),"'style' must be string"
		assert isinstance(_size, int), "'size' must be integer"
		assert _size>0, "size>0 expected"

		self.style = _style
		self.size = _size
		self.Pen = Pen(kwargs)
		self.Brush = Brush(kwargs)

	
	def __iter__(self):
		return iter([
			("style",self.style),
			("size", self.size),
			("fill", dict(iter(self.Brush))),
			("line", dict(iter(self.Pen)))
		])


""" Food Process Engineering library (FPE) """


""" Food is the base class"""
from .foodmaterial import Food, Beverage, Juice, Cereal, Legume, Nut, Dairy, Fruit, Vegetable, Meat, Sweet


"""Different methods to compute water activity and specific heat capacity"""
from .foodproperty import Aw, Cp


import numpy as _np
import ctypes as _ct
from typing import Iterable
from numbers import Real
from .._ctypeslib import pydll as _pydll



# ----- Standard Beta Distribution  -------

def dbeta(x:Iterable|Real, shape1:float, shape2:float)->list|Real:
	"""
	shape1, shape2: similar to alpha and beta
	"""
	assert shape1>0, "shape1>0 expected"
	assert shape2>0, "shape2>0 expected"
	return _pydll.c_stat_dbeta(_ct.py_object(x), _ct.c_double(shape1), _ct.c_double(shape2))


def pbeta(q:Iterable|Real, shape1:float, shape2:float)->list|Real:
	"""
	shape1, shape2: similar to alpha and beta
	"""
	assert shape1>0, "shape1>0 expected"
	assert shape2>0, "shape2>0 expected"

	return _pydll.c_stat_pbeta(_ct.py_object(q), _ct.c_double(shape1), _ct.c_double(shape2))


def qbeta(p:Iterable|Real, shape1:float, shape2:float)->list|Real:
	"""
	shape1, shape2: similar to alpha and beta
	"""
	assert shape1>0, "shape1>0 expected"
	assert shape2>0, "shape2>0 expected"

	return _pydll.c_stat_qbeta(_ct.py_object(p), _ct.c_double(shape1), _ct.c_double(shape2))
	

def rbeta(n:int, shape1:float, shape2:float)->list:
	"""
	shape1, shape2: similar to alpha and beta
	"""
	assert n>0 ,"n>0 expected"
	assert shape1>0, "shape1>0 expected"
	assert shape2>0, "shape2>0 expected"

	return _np.random.beta(size=n, a=shape1, b=shape2).tolist()



# ----- Binomial Distribution  -------

def dbinom(x:Iterable|Real, size:int, prob:float)->list|Real:
	"""
	size: number of trials
	prob: probability of success in each trial
	"""
	return _pydll.c_stat_dbinom(_ct.py_object(x), _ct.c_int(size), _ct.c_double(prob))


def pbinom(q:Iterable|Real, size:int, prob:float)->list|Real:
	"""
	size: number of trials
	prob: probability of success in each trial
	"""
	assert size>0, "size>0 expected"
	assert prob>=0 and prob<=1, "prob in [0, 1] expected"

	return _pydll.c_stat_pbinom(_ct.py_object(q), _ct.c_int(size), _ct.c_double(prob))


def qbinom(p:Iterable|Real, size:int, prob:float)->list|Real:
	"""
	size: number of trials
	prob: probability of success in each trial
	"""
	assert size>0, "size>0 expected"
	assert prob>=0 and prob<=1, "prob in [0, 1] expected"

	return _pydll.c_stat_qbinom(_ct.py_object(p), _ct.c_int(size), _ct.c_double(prob))
	

def rbinom(n:int, size:int, prob:float)->list:
	"""
	size: number of trials
	prob: probability of success in each trial
	"""
	assert n>0 ,"n>0 expected"
	assert size>0, "size>0 expected"
	assert prob>=0 and prob<=1, "prob in [0, 1] expected"

	return _np.random.binomial(size=n, n=size, p=prob).tolist()




# ----- Negative-Binomial Distribution  -------

def dnbinom(x:Iterable|Real, size:int, prob:float)->list|Real:
	"""
	x: quantiles representing number of failures
	size: target for number of successful trials
	prob: probability of success in each trial
	"""
	assert size>0, "size>0 expected"
	assert prob>=0 and prob<=1, "prob in [0, 1] expected"

	return _pydll.c_stat_dnbinom(_ct.py_object(x), _ct.c_int(size), _ct.c_double(prob))


def pnbinom(q:Iterable|Real, size:int, prob:float)->list|Real:
	"""
	q: quantiles representing number of failures
	size: target for number of successful trials
	prob: probability of success in each trial
	"""
	assert size>0, "size>0 expected"
	assert prob>=0 and prob<=1, "prob in [0, 1] expected"

	return _pydll.c_stat_pnbinom(_ct.py_object(q), _ct.c_int(size), _ct.c_double(prob))

def qnbinom(p:Iterable|Real, size:int, prob:float)->list|Real:
	"""
	p: probabilities
	size: target for number of successful trials
	prob: probability of success in each trial
	"""
	assert size>0, "size>0 expected"
	assert prob>=0 and prob<=1, "prob in [0, 1] expected"

	return _pydll.c_stat_qnbinom(_ct.py_object(p), _ct.c_int(size), _ct.c_double(prob))


def rnbinom(n:int, size, prob)->list:
	"""
	size: target for number of successful trials
	prob: probability of success in each trial
	"""
	assert n>0 ,"n>0 expected"
	assert size>0, "size>0 expected"
	assert prob>=0 and prob<=1, "prob in [0, 1] expected"

	return _np.random.negative_binomial(size=n, n=size, p=prob).tolist()



# ----- Multinomial Distribution  -------

def dmultinom(x:Iterable, size:int, prob:Iterable)->float:
	"""
	x: quantiles 
	size: number of trials
	prob: probabilities of success in each trial

	## Note:
	Internally sum of probabilities is normalized to 1.0
	"""

	assert sum(x) == size, "sum(x) == size expected."
	assert size>0, "size>0 expected"

	return _pydll.c_stat_dmultinom(_ct.py_object(x), _ct.c_int(size), prob)


def rmultinom(n:int, size:int, prob:Iterable)->list:
	"""
	size: number of trials
	prob: probabilities of success in each trial

	returns 2D list, where list[i] corresponds to prob[i]

	## Note:
	Sum of probabilities is normalized to 1.0
	"""
	assert size>0, "size>0 expected"
	assert len(prob)>1, "For single prob value use rbinom function"

	Sum = sum(prob) 
	assert Sum>0, "sum of probabilities must be >0"

	return [rbinom(n, size, p/Sum) for p in prob]




# ----- Chi-Square Distribution  -------

def dchisq(x:Iterable|Real, df:int)->list|Real:
	"""
	df: degrees of freedom
	"""
	assert df>0, "df>0 expected"
	return _pydll.c_stat_dchisq(_ct.py_object(x), _ct.c_int(df))


def pchisq(q:Iterable|Real, df:int)->list|Real:
	"""
	df: degrees of freedom
	"""
	assert df>0, "df>0 expected"	
	return _pydll.c_stat_pchisq(_ct.py_object(q), _ct.c_int(df))


def qchisq(p:Iterable|Real, df:int)->list|Real:
	"""
	df: degrees of freedom
	"""
	assert df>0, "df>0 expected"
	return _pydll.c_stat_qchisq(_ct.py_object(p), _ct.c_int(df))


def rchisq(n:int, df)->list:
	"""
	df: degrees of freedom
	"""
	assert n>0 ,"n>0 expected"
	assert df>0, "df>0 expected"

	return _np.random.chisquare(size=n, df = df).tolist()



# ----- Exponential Distribution  -------

def dexp(x:Iterable|Real, rate = 1.0)->list|Real:
	"""
	x: quantiles
	rate: 1/mean, where mean is the waiting time for the next event recurrence
	"""
	return _pydll.c_stat_dexp(_ct.py_object(x), _ct.c_double(rate))


def pexp(q:Iterable|Real, rate = 1.0)->list|Real:
	"""
	q: quantiles
	rate: 1/mean, where mean is the waiting time for the next event recurrence
	"""
	return _pydll.c_stat_pexp(_ct.py_object(q),  _ct.c_double(rate))


def qexp(p:Iterable|Real, rate = 1.0)->list|Real:
	"""
	p: probabilities
	rate: 1/mean, where mean is the waiting time for the next event recurrence
	"""
	return _pydll.c_stat_qexp(_ct.py_object(p), _ct.c_double(rate))


def rexp(n:int, rate=1.0)->list:
	"""
	rate: 1/mean, where mean is the waiting time for the next event recurrence
	"""
	assert n>0 ,"n>0 expected"
	assert rate>0, "rate>0 expected"

	return _np.random.exponential(size=n, scale=1/rate).tolist()



# ----- F Distribution  -------

def df(x:Iterable|Real, df1:int, df2:int)->list|Real:
	"""
	df1: degrees of freedom, numerator
	df2: degrees of freedom, denominator
	"""
	assert df1>0, "df1>0 expected"
	assert df2>0, "df2>0 expected"

	return _pydll.c_stat_df(_ct.py_object(x), _ct.c_int(df1), _ct.c_int(df2))


def pf(q:Iterable|Real, df1:int, df2:int)->list|Real:
	"""
	df1: degrees of freedom, numerator
	df2: degrees of freedom, denominator
	"""
	assert df1>0, "df1>0 expected"
	assert df2>0, "df2>0 expected"

	return _pydll.c_stat_pf(_ct.py_object(q), _ct.c_int(df1), _ct.c_int(df2))

def qf(p:Iterable|Real, df1:int, df2:int)->list|Real:
	"""
	df1: degrees of freedom, numerator
	df2: degrees of freedom, denominator
	"""
	assert df1>0, "df1>0 expected"
	assert df2>0, "df2>0 expected"

	return _pydll.c_stat_qf(_ct.py_object(p), _ct.c_int(df1), _ct.c_int(df2))


def rf(n:int, df1, df2)->list:
	"""
	df1: degrees of freedom, numerator
	df2: degrees of freedom, denominator
	"""
	assert n>0 ,"n>0 expected"
	assert df1>0, "df1>0 expected"
	assert df2>0, "df2>0 expected"

	return _np.random.f(size=n, dfnum=df1, dfden=df2).tolist()




# ----- Gamma Distribution  -------

def dgamma(x:Iterable|Real, shape:Real, scale = 1.0)->list|Real:
	"""
	x: quantile	
	shape: waiting time for the rth event to occur
	scale: average waiting time for the next event recurrence
	"""
	return _pydll.c_stat_dgamma(_ct.py_object(x), _ct.c_double(shape), _ct.c_double(scale))


def pgamma(q:Iterable|Real, shape:float, scale = 1.0)->list|Real:
	"""
	q: quantile
	shape: waiting time for the rth event to occur
	scale: average waiting time for the next event recurrence
	"""
	return _pydll.c_stat_pgamma(_ct.py_object(q), _ct.c_double(shape), _ct.c_double(scale))


def qgamma(p:Iterable|Real, shape:Real, scale = 1.0)->list|Real:
	"""
	p: probabilities
	shape: waiting time for the rth event to occur
	scale: average waiting time for the next event recurrence
	"""
	return _pydll.c_stat_qgamma(_ct.py_object(p), _ct.c_double(shape), _ct.c_double(scale))


def rgamma(n:int, shape:Real, scale=1.0)->list:
	"""
	Draw samples from gamma distribution
	"""
	assert n>0 ,"n>0 expected"
	assert shape>0, "shape>0 expected"
	assert scale>0, "scale>0 expected"

	return _np.random.gamma(size=n, scale=scale, shape=shape).tolist()




# ----- Geometric Distribution  -------

def dgeom(x:Iterable|Real, prob:float)->list|Real:
	"""
	x: Number of failures before success occurs.
	prob: probability of success in each trial.
	"""
	return _pydll.c_stat_dgeom(_ct.py_object(x), _ct.c_double(prob))


def pgeom(q:Iterable|Real, prob:Real)->list|Real:
	"""
	q: Number of failures before success occurs.
	prob: probability of success in each trial.
	"""
	return _pydll.c_stat_pgeom(_ct.py_object(q), _ct.c_double(prob))


def qgeom(p:Iterable|Real, prob:Real)->list|Real:
	"""
	p: probabilities
	prob: probability of success in each trial.
	"""
	return _pydll.c_stat_qgeom(_ct.py_object(p), _ct.c_double(prob))


def rgeom(n:int, prob:Real)->list:
	"""
	Draw samples from geometric distribution
	"""
	assert n>0 ,"n>0 expected"
	assert prob>=0 and prob<=1, "prob in [0, 1] expected"

	return _np.random.geometric(size=n, p=prob).tolist()




# ----- Hypergeometric Distribution  -------

def dhyper(x:Iterable|Real, m:int, n:int, k:int)->list|Real:
	"""
	m: number of good samples in the urn
	n: number of bad samples in the urn
	k: samples drawn from the urn
	"""
	return _pydll.c_stat_dhyper(_ct.py_object(x), _ct.c_int(m), _ct.c_int(n), _ct.c_int(k))


def phyper(q:Iterable|Real, m:int, n:int, k:int)->list|Real:
	"""
	m: number of good samples in the urn
	n: number of bad samples in the urn
	k: samples drawn from the urn
	"""
	return _pydll.c_stat_phyper(_ct.py_object(q), _ct.c_int(m), _ct.c_int(n), _ct.c_int(k))


def qhyper(p:Iterable|Real, m:int, n:int, k:int)->list|Real:
	"""
	m: number of good samples in the urn
	n: number of bad samples in the urn
	k: samples drawn from the urn
	"""
	return _pydll.c_stat_qhyper(_ct.py_object(p), _ct.c_int(m), _ct.c_int(n), _ct.c_int(k))


def rhyper(nn:int, m:int, n:int, k:int)->list:
	"""
	m: number of good samples in the urn
	n: number of bad samples in the urn
	k: samples drawn from the urn
	"""
	assert nn>0 ,"nn>0 expected"
	assert m>0, "m>0 expected"
	assert n>0, "n>0 expected"
	assert k>0, "k>0 expected"

	return _np.random.hypergeometric(size=nn, ngood=m, nbad=n, nsample=k)




# ----- Normal Distribution  -------

def dnorm(x:Iterable|Real, mean=0.0, sd=1.0)->list|Real:
	"""
	mean: mean value of the distribution
	sd: standard deviation of the distribution
	"""
	return _pydll.c_stat_dnorm(_ct.py_object(x), _ct.c_double(mean), _ct.c_double(sd))


def pnorm(q:Iterable|Real, mean=0.0, sd=1.0)->list|Real:
	"""
	mean: mean value of the distribution
	sd: standard deviation of the distribution
	"""
	return _pydll.c_stat_pnorm(_ct.py_object(q), _ct.c_double(mean), _ct.c_double(sd))


def qnorm(p:Iterable|Real, mean=0.0, sd=1.0)->list|Real:
	"""
	mean: mean value of the distribution
	sd: standard deviation of the distribution
	"""
	return _pydll.c_stat_qnorm(_ct.py_object(p), _ct.c_double(mean), _ct.c_double(sd))


def rnorm(n:int, mean=0.0, sd=1.0)->list:
	"""
	mean: mean value of the distribution
	sd: standard deviation of the distribution
	"""
	assert n>0 ,"n>0 expected"
	assert sd>0, "sd>0 expected"

	return _np.random.normal(size=n, loc=mean, scale=sd).tolist()




# ----- Log Normal Distribution  -------

def dlnorm(x:Iterable|Real, meanlog=0.0, sdlog=1.0)->list|Real:
	"""
	meanlog: mean value of the distribution
	sdlog: standard deviation of the distribution
	"""
	return _pydll.c_stat_dlnorm(_ct.py_object(x), _ct.c_double(meanlog), _ct.c_double(sdlog))


def plnorm(q:Iterable|Real, meanlog=0.0, sdlog=1.0)->list|Real:
	"""
	mean: mean value of the distribution
	sd: standard deviation of the distribution
	"""
	return _pydll.c_stat_plnorm(_ct.py_object(q), _ct.c_double(meanlog), _ct.c_double(sdlog))


def qlnorm(p:Iterable|Real, meanlog=0.0, sdlog=1.0)->list|Real:
	"""
	mean: mean value of the distribution
	sd: standard deviation of the distribution
	"""
	return _pydll.c_stat_qlnorm(_ct.py_object(p), _ct.c_double(meanlog), _ct.c_double(sdlog))


def rlnorm(n:int, meanlog=0.0, sdlog=1.0)->list:
	"""
	mean: mean value of the distribution
	sd: standard deviation of the distribution
	"""
	assert n>0 ,"n>0 expected"
	assert sdlog>0, "sd>0 expected"

	return _np.random.lognormal(size=n, mean=meanlog, sigma=sdlog).tolist()




# ----- Pareto Distribution  -------

def dpareto(x:Iterable|Real, location:Real, shape=1.0)->list|Real:
	"""
	location: location parameter
	shape: shape parameter
	"""
	assert location>0 and shape>0, "'location' and 'shape' must be positive"
	return _pydll.c_stat_dpareto(_ct.py_object(x), _ct.c_double(location), _ct.c_double(shape))


def ppareto(q:Iterable|Real, location:Real, shape=1.0)->list|Real:
	"""
	location: location parameter
	shape: shape parameter
	"""
	assert location>0 and shape>0, "'location' and 'shape' must be positive"
	return _pydll.c_stat_ppareto(_ct.py_object(q), _ct.c_double(location), _ct.c_double(shape))


def qpareto(p:Iterable|Real, location:Real, shape=1.0)->list|Real:
	"""
	location: location parameter
	shape: shape parameter
	"""
	assert location>0 and shape>0, "'location' and 'shape' must be positive"
	return _pydll.c_stat_qpareto(_ct.py_object(p), _ct.c_double(location), _ct.c_double(shape))


def rpareto(n:int, location:Real, shape=1.0)->list:
	"""
	location: location parameter
	shape: shape parameter
	"""
	assert location>0 and shape>0, "'location' and 'shape' must be positive"
	assert n>0 ,"n>0 expected"
	
	return (_np.random.pareto(size=n, a=shape)*location).tolist()




# ----- Poisson Distribution  -------

def dpois(x:Iterable|Real, mu:float)->list|Real:
	return _pydll.c_stat_dpois(_ct.py_object(x), _ct.c_double(mu))


def ppois(q:Iterable|Real, mu:float)->list|Real:
	return _pydll.c_stat_ppois(_ct.py_object(q), _ct.c_double(mu))


def qpois(p:Iterable|Real, mu:float)->list|Real:
	return _pydll.c_stat_qpois(_ct.py_object(p), _ct.c_double(mu))


def rpois(n:int, mu = 1)->list:
	"""
	Draw samples from Poisson distribution
	"""
	assert n>0 ,"n>0 expected"
	assert mu>0, "mu>0 expected"

	return _np.random.poisson(size=n, lam=mu).tolist()




# ----- t Distribution  -------

def dt(x:Iterable|Real, df:int)->list|Real:
	"""
	df: degrees of freedom
	"""
	assert df>0, "df>0 expected"

	return _pydll.c_stat_dt(_ct.py_object(x), _ct.c_int(df))


def pt(q:Iterable|Real, df:int)->list|Real:
	"""
	df: degrees of freedom
	"""
	assert df>0, "df>0 expected"

	return _pydll.c_stat_pt(_ct.py_object(q), _ct.c_int(df))


def qt(p:Iterable|Real, df:int)->list|Real:
	"""
	df: degrees of freedom
	"""
	assert df>0, "df>0 expected"

	return _pydll.c_stat_qt(_ct.py_object(p), _ct.c_int(df))


def rt(n:int, df)->list:
	"""
	df: degrees of freedom
	"""
	assert n>0 ,"n>0 expected"
	assert df>0, "df1>0 expected"

	return _np.random.standard_t(size=n, df=df).tolist()




# ----- Uniform Distribution  -------

def dunif(x:Iterable|Real, min=0.0, max=1.0)->list|Real:
	"""
	min: minimum bound
	max: maximum bound
	"""
	assert max>min, "max>min expected"

	return _pydll.c_stat_dunif(_ct.py_object(x), _ct.c_double(min), _ct.c_double(max))


def punif(q:Iterable|Real, min=0.0, max=1.0)->list|Real:
	"""
	min: minimum bound
	max: maximum bound
	"""
	assert max>min, "max>min expected"

	return _pydll.c_stat_punif(_ct.py_object(q), _ct.c_double(min), _ct.c_double(max))


def qunif(p:Iterable|Real, min=0.0, max=1.0)->list|Real:
	"""
	min: minimum bound
	max: maximum bound
	"""
	assert max>min, "max>min expected"

	return _pydll.c_stat_qunif(_ct.py_object(p), _ct.c_double(min), _ct.c_double(max))


def runif(n:int, min=0.0, max=1.0)->list:
	"""
	min: minimum bound
	max: maximum bound
	"""
	assert max>min, "max>min expected"
	assert n>0 ,"n>0 expected"

	return _np.random.uniform(size=n, low=min, high=max).tolist()



# ----- Weibull Distribution  -------

def dweibull(x:Iterable|Real, shape:float, scale = 1.0)->list|Real:
	"""
	x: quantile	
	shape: known as Weibull-slope
	scale: characteristic life
	"""
	assert shape>0, "shape>0 expected"
	assert scale>0, "scale>0 expected"

	return _pydll.c_stat_dweibull(_ct.py_object(x), _ct.c_double(shape), _ct.c_double(scale))


def pweibull(q:Iterable|Real, shape:float, scale = 1.0)->list|Real:
	"""
	q: quantile
	shape: known as Weibull-slope
	scale: characteristic life
	"""
	assert shape>0, "shape>0 expected"
	assert scale>0, "scale>0 expected"

	return _pydll.c_stat_pweibull(_ct.py_object(q), _ct.c_double(shape), _ct.c_double(scale))


def qweibull(p:Iterable|Real, shape:float, scale = 1.0)->list|Real:
	"""
	p: probabilities
	shape: known as Weibull-slope
	scale: characteristic life
	"""
	assert shape>0, "shape>0 expected"
	assert scale>0, "scale>0 expected"
	
	return _pydll.c_stat_qweibull(_ct.py_object(p), _ct.c_double(shape), _ct.c_double(scale))


def rweibull(n:int, shape:float, scale=1.0)->list:
	"""
	Draw samples from weibull distribution
	"""
	assert n>0 ,"n>0 expected"
	assert shape>0, "shape>0 expected"
	assert scale>0, "scale>0 expected"

	return (scale*_np.random.weibull(size=n, a=shape, )).tolist()


# ----- Wilcoxon Sign Rank Distribution  -------

def dsignrank(x:Iterable|Real, n:int)->list|Real:
	return _pydll.c_stat_dsignrank(_ct.py_object(x), _ct.c_int(n))


def psignrank(q:Iterable|Real, n:int)->list|Real:
	return _pydll.c_stat_psignrank(_ct.py_object(q), _ct.c_int(n))


def qsignrank(p:Iterable|Real, n:int)->list|Real:
	return _pydll.c_stat_qsignrank(_ct.py_object(p), _ct.c_int(n))




import numpy as _np
from typing import Iterable as _Iterable
from ..stats import qnorm
from .charts import scatter
from .chartelems import Marker
from ..util import minmax
from ..fitting import approx

def _ComputeQuantiles(X:_Iterable)->_Iterable:
	lenX = len(X)
	Quantiles = [0.0]*lenX

	a = 0.375 if lenX <= 10 else 0.5
	Percentile = lambda Order: (Order - a)/(lenX + 1.0 - 2*a)
	
	i=1
	for j in range(lenX):
		Percent = Percentile(i)
		Quantiles[j] = qnorm(Percent)
		i += 1
	
	return Quantiles



def qqnorm(
		data:_Iterable, 
		label:str=None, 
		show=True, 
		marker:str|Marker=None,
		**kwargs):
	"""
	Normal Quantile-quantile chart \n
	x-axis="Theoretical Quantiles" \n  
	y-axis="Sample Quantiles" \n

	## Input:
	data: Data \n
	show: Whether to show theoretical line or not 
	"""
	assert isinstance(data, _Iterable), "'data' must be iterable"
	if label!=None:
		assert isinstance(label, str), "'label' must be string"
	
	assert isinstance(show, bool), "'show' must be bool"

	_mark = marker or Marker()
	assert isinstance(_mark, str|Marker), "marker must be str|Marker"
	if isinstance(_mark, str):
		_mark = Marker(style=marker, size=kwargs.get("markersize") or 5)
	

	Quantiles = _ComputeQuantiles(data)

	if show:		
		MinQ, MaxQ = minmax(Quantiles)

		#Extend the line by 10% in both directions
		MinQ = MinQ*1.1
		MaxQ = MaxQ*1.1

		lwidth = (kwargs.get("lw") or kwargs.get("linewidth")) or 2
		scatter(x=(MinQ, MaxQ), y=(MinQ, MaxQ), marker=None, **kwargs)

	scatter(x=Quantiles, y=_np.sort(data), label=label, marker=marker)





def qqplot(
		x:_Iterable, 
		y:_Iterable, 
		**kwargs):
	"""
	Plots quantile-quantile chart using two data-sets (x,y)

	## Input
	x, y: Data
	"""
	assert isinstance(x, _Iterable), "'x' must be iterable"
	assert isinstance(y, _Iterable), "'y' must be iterable"

	xx = _np.array(x)
	xx.sort()

	yy = _np.array(y)
	yy.sort()

	lenX, lenY = len(xx), len(yy)

	if lenX != lenY:
		InterpData = yy if lenY>lenX else xx
		xdata = _np.arange(1, len(InterpData)+1)

		InterpPoints = approx(xdata, InterpData, min(lenX, lenY))

		if lenY>lenX:
			yy=InterpPoints[1]
		else:
			xx=InterpPoints[1]
	
	_style = kwargs.get("marker") or "c"
	_size = kwargs.get("markersize") or 5
	marker = Marker(style=_style, size=_size, **kwargs)

	scatter(x=xx, y=yy, marker=marker)


